# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import shutil
import unittest
import logging
from datetime import datetime, time, date

from geopathfinder.naming_conventions.yeoda_naming import YeodaFilename
from geopathfinder.naming_conventions.yeoda_naming import yeoda_path
from geopathfinder.naming_conventions.yeoda_naming import yeoda_tree

logging.basicConfig(level=logging.INFO)


class TestYeodaFilename(unittest.TestCase):

    def setUp(self):
        self.dtime_1 = datetime(2008, 1, 1, 12, 23, 33)
        self.dtime_2 = datetime(2009, 2, 2, 22, 1, 1)

        fields = {'datetime_1': self.dtime_1, 'datetime_2': self.dtime_2, 'var_name': 'SSM'}

        self.yeoda_fn = YeodaFilename(fields, convert=True)

        fields = {'datetime_1': self.dtime_1, 'var_name': 'SSM', 'band': 'VV'}

        self.yeoda_fn2 = YeodaFilename(fields, convert=True)

        self.yeoda_fn3 = YeodaFilename(fields)

        fn = 'SIG0_20170725T165004__VV_A146_E048N012T6_EU500M_V04R01_S1BIWG1.tif'
        self.yeoda_fn4 = YeodaFilename.from_filename(fn, convert=True)

        fn = 'TMENSIG40_20170725_20181225__A146_E048N012T6_EU500M__ASAWS.tif'
        self.yeoda_fn5 = YeodaFilename.from_filename(fn)

        fn = 'TMENSIG40_20170725_20181225_M1__E048N012T6_EU500M_V04R01_ASAWS.tif'
        self.yeoda_fn6 = YeodaFilename.from_filename(fn, convert=True)

        fn = 'SIG0_20210128T184253__VH__E036N039T3_EU020M_V1M0R1_S1AIWGRDH_TUWIEN.tif'
        self.yeoda_fn7a = YeodaFilename.from_filename(fn, convert=True)
        fn = 'SIG0_20210128T184253__VH__E036N039T3_EU020M_V1M0R1_S1AIWGRDH.tif'
        self.yeoda_fn7b = YeodaFilename.from_filename(fn, convert=True)


    def test1_build_yeoda_filename(self):
        """
        Test building Yeoda file name.

        """
        fn = ('SSM_20080101T122333_20090202T220101______.tif')

        self.assertEqual(str(self.yeoda_fn), fn)


    def test2_get_n_set_date(self):
        """
        Test set and get start and end date.

        """
        self.assertEqual(self.yeoda_fn['datetime_1'], self.dtime_1)
        self.assertEqual(self.yeoda_fn['datetime_2'], self.dtime_2)

        new_start_time = datetime(2009, 1, 1, 12, 23, 33)
        self.yeoda_fn['datetime_1'] = new_start_time

        self.assertEqual(self.yeoda_fn['datetime_1'], new_start_time)
        self.assertEqual(self.yeoda_fn['datetime_2'], self.dtime_2)

    def test30_get_n_set_date_n_time(self):
        """
        Test set and get date and time for a single datetime.

        """
        self.assertEqual(self.yeoda_fn2['datetime_1'].date(), self.dtime_1.date())
        self.assertEqual(self.yeoda_fn2['datetime_1'].time(), self.dtime_1.time())

        new_start_time = datetime(2009, 1, 1, 12, 23, 33)
        self.yeoda_fn2['datetime_1'] = new_start_time

        self.assertEqual(self.yeoda_fn2['datetime_1'].date(), new_start_time.date())
        self.assertEqual(self.yeoda_fn2['datetime_1'].time(), new_start_time.time())

    def test31_get_n_set_date_n_time_dts(self):
        """
        Test get and set date and time for a single datetime,
        returning strings.

        """
        self.assertEqual(self.yeoda_fn2.obj.datetime_1.date(), datetime(2008, 1, 1).date())
        self.assertEqual(self.yeoda_fn2.obj.datetime_1.time(), time(12, 23, 33))

        new_start_time = date(2345, 1, 2)
        self.yeoda_fn2['datetime_1'] = new_start_time

        self.assertEqual(self.yeoda_fn2.obj.datetime_1.time(), time())

    def test32_get_n_set_date_n_time_strings(self):
        """
        Test get and set date and time for a single datetime,
        returning strings.

        """

        self.assertEqual(self.yeoda_fn3.obj.datetime_1, '20080101T122333')

        new_start_time = datetime(2345, 1, 2, 7, 8, 9)
        self.yeoda_fn3['datetime_2'] = new_start_time

        self.assertEqual(self.yeoda_fn3.obj.datetime_1, '20080101T122333')
        self.assertEqual(self.yeoda_fn3.obj.datetime_2, '23450102T070809')

    def test4_create_yeoda_filename(self):
        """
        Tests the creation of a SmartFilename from a given string filename.

        """

        # testing for single datetime
        self.assertEqual(self.yeoda_fn4['datetime_1'].date(), datetime(2017, 7, 25).date())
        self.assertEqual(self.yeoda_fn4['datetime_1'].time(), time(16, 50, 4))
        self.assertEqual(self.yeoda_fn4['var_name'], 'SIG0')
        self.assertEqual(self.yeoda_fn4['sensor_field'], 'S1BIWG1')
        self.assertEqual(self.yeoda_fn4['band'], 'VV')
        self.assertEqual(self.yeoda_fn4['extra_field'], 'A146')
        self.assertEqual(self.yeoda_fn4['data_version'], 'V04R01')
        self.assertEqual(self.yeoda_fn4['grid_name'], 'EU500M')
        self.assertEqual(self.yeoda_fn4['tile_name'], 'E048N012T6')
        self.assertEqual(self.yeoda_fn4.ext, '.tif')

        # testing for empty fields and two dates
        self.assertEqual(self.yeoda_fn5['datetime_1'], '20170725')
        self.assertEqual(self.yeoda_fn5['datetime_2'], '20181225')
        self.assertEqual(self.yeoda_fn5['var_name'], 'TMENSIG40')
        self.assertEqual(self.yeoda_fn5['sensor_field'], 'ASAWS')
        self.assertEqual(self.yeoda_fn5['band'], '')
        self.assertEqual(self.yeoda_fn5['data_version'], '')

        # testing for empty relative orbit field
        self.assertEqual(self.yeoda_fn6['extra_field'], None)

    def test5_build_ascat_ssm_fname(self):
        """
        Tests the creation of a resampled ASCAT SSM filename.

        """
        date_time = '20331122_112233'
        tilename = 'EU500M_E012N024T6'

        xfields = {'datetime_1': datetime.strptime(date_time, "%Y%m%d_%H%M%S"),
                   'var_name': 'SSM',
                   'sensor_field': 'ASCSMO12NA',
                   'band': 'XX',
                   'extra_field': 'D',
                   'data_version': 'V2M3R1',
                   'grid_name': tilename[:6],
                   'tile_name': tilename[7:]}

        should = 'SSM_20331122T112233__XX_D_E012N024T6_EU500M_V2M3R1_ASCSMO12NA.tif'
        fn = YeodaFilename(xfields)
        self.assertEqual(str(fn), should)

    def test6_compact_fname(self):
        """
        Tests if the compact representation of the filename works as expected.

        """
        self.yeoda_fn4['data_version'] = 'R01'
        fn_str = str(self.yeoda_fn4)
        self.assertEqual(fn_str, 'SIG0_20170725T165004__VV_A146_E048N012T6_EU500M_R01_S1BIWG1.tif')

        self.yeoda_fn['data_version'] = 'V01R0'
        fn_str = str(self.yeoda_fn)
        self.assertEqual(fn_str, 'SSM_20080101T122333_20090202T220101_____V01R0_.tif')

        self.yeoda_fn['data_version'] = 'V1M3R3'
        fn_str = str(self.yeoda_fn)
        self.assertEqual(fn_str, 'SSM_20080101T122333_20090202T220101_____V1M3R3_.tif')

        self.yeoda_fn['data_version'] = 'V2R4'
        fn_str = str(self.yeoda_fn)
        self.assertEqual(fn_str, 'SSM_20080101T122333_20090202T220101_____V2R4_.tif')

        self.yeoda_fn['data_version'] = ''
        fn_str = str(self.yeoda_fn)
        self.assertEqual(fn_str, 'SSM_20080101T122333_20090202T220101______.tif')

        self.yeoda_fn['data_version'] = 'V001R002'
        fn_str = str(self.yeoda_fn)
        self.assertEqual(fn_str, 'SSM_20080101T122333_20090202T220101_____V001R002_.tif')

    def test7_replace_fnparts(self):
        """
        Test all types of fields can be altered after creating the SmartFilename object.

        """
        self.yeoda_fn5['data_version'] = 'V01R01'
        self.assertEqual(str(self.yeoda_fn5), 'TMENSIG40_20170725_20181225__A146_E048N012T6_EU500M_V01R01_ASAWS.tif')
        self.assertEqual(self.yeoda_fn5['data_version'], 'V01R01')

        self.yeoda_fn5['data_version'] = 'V1M2R1'
        self.assertEqual(str(self.yeoda_fn5), 'TMENSIG40_20170725_20181225__A146_E048N012T6_EU500M_V1M2R1_ASAWS.tif')

        self.yeoda_fn5['data_version'] = ''
        self.yeoda_fn5['band'] = 'VV'
        self.assertEqual(str(self.yeoda_fn5), 'TMENSIG40_20170725_20181225_VV_A146_E048N012T6_EU500M__ASAWS.tif')
        self.assertEqual(self.yeoda_fn5['band'], 'VV')

        self.yeoda_fn5['band'] = ''
        self.assertEqual(str(self.yeoda_fn5), 'TMENSIG40_20170725_20181225__A146_E048N012T6_EU500M__ASAWS.tif')


    def test8_optional_last_field(self):

        # testing for case with creator field
        self.assertEqual(self.yeoda_fn7a._build_fn(), 'SIG0_20210128T184253__VH__E036N039T3_EU020M_V1M0R1_S1AIWGRDH_TUWIEN.tif')

        self.assertEqual(self.yeoda_fn7a['datetime_1'].date(), datetime(2021, 1, 28).date())
        self.assertEqual(self.yeoda_fn7a['datetime_1'].time(), time(18, 42, 53))
        self.assertEqual(self.yeoda_fn7a['var_name'], 'SIG0')
        self.assertEqual(self.yeoda_fn7a['sensor_field'], 'S1AIWGRDH')
        self.assertEqual(self.yeoda_fn7a['band'], 'VH')
        self.assertEqual(self.yeoda_fn7a['extra_field'], None)
        self.assertEqual(self.yeoda_fn7a['data_version'], 'V1M0R1')
        self.assertEqual(self.yeoda_fn7a['grid_name'], 'EU020M')
        self.assertEqual(self.yeoda_fn7a['tile_name'], 'E036N039T3')
        self.assertEqual(self.yeoda_fn7a['creator'], 'TUWIEN')
        self.assertEqual(self.yeoda_fn7a.ext, '.tif')

        # testing for case without creator field
        self.assertEqual(self.yeoda_fn7b._build_fn(), 'SIG0_20210128T184253__VH__E036N039T3_EU020M_V1M0R1_S1AIWGRDH.tif')

        self.assertEqual(self.yeoda_fn7b['datetime_1'].date(), datetime(2021, 1, 28).date())
        self.assertEqual(self.yeoda_fn7b['datetime_1'].time(), time(18, 42, 53))
        self.assertEqual(self.yeoda_fn7b['var_name'], 'SIG0')
        self.assertEqual(self.yeoda_fn7b['sensor_field'], 'S1AIWGRDH')
        self.assertEqual(self.yeoda_fn7b['band'], 'VH')
        self.assertEqual(self.yeoda_fn7b['extra_field'], None)
        self.assertEqual(self.yeoda_fn7b['data_version'], 'V1M0R1')
        self.assertEqual(self.yeoda_fn7b['grid_name'], 'EU020M')
        self.assertEqual(self.yeoda_fn7b['tile_name'], 'E036N039T3')
        self.assertEqual(self.yeoda_fn7b['creator'], None)
        self.assertEqual(self.yeoda_fn7b.ext, '.tif')


class TestYeodaPath(unittest.TestCase):
    """
    Tests checking if a yeoda path is correctly reflected by yeoda_tree.
    """

    def setUp(self):
        """
        Setting up the test yeoda_tree.
        """

        self.test_dir = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), 'test_data', 'Sentinel-1_CSAR_IWGRDH')

    def test_full_path(self):
        """
        Tests the SmartPath() for the yeoda naming conventions
        """

        # passing the version as string directly
        should = os.path.join(self.test_dir, 'SSM', 'V3M2R1', 'EQUI7_EU500M', 'E048N012T6', 'qlooks')
        stp1 = yeoda_path(self.test_dir, product='SSM', data_version='V3M2R1', grid='EQUI7_EU500M',
                          tile='E048N012T6', qlook=True, make_dir=False)

        self.assertEqual(stp1.directory, should)

        # passing the version as integer
        should = os.path.join(self.test_dir, 'SSM', 'V7', 'EQUI7_EU500M', 'E048N012T6', 'qlooks')
        stp2 = yeoda_path(self.test_dir, product='SSM', data_version=7, grid='EQUI7_EU500M',
                          tile='E048N012T6', qlook=True, make_dir=False)
        self.assertEqual(stp2.directory, should)

        should = os.path.join(self.test_dir, 'SSM', 'V0', 'EQUI7_EU500M', 'E048N012T6', 'qlooks')
        stp3 = yeoda_path(self.test_dir, product='SSM', data_version=0, grid='EQUI7_EU500M',
                          tile='E048N012T6', qlook=True, make_dir=False)
        self.assertEqual(stp3.directory, should)


    def test_data_version(self):
        """
        Tests if the dataversion can be handled correctly.

        """
        stp1 = yeoda_path(self.test_dir, product='SSM', data_version='V3M2R1', grid='EQUI7_EU500M',
                          tile='E048N012T6', qlook=True, make_dir=False)
        self.assertEqual(stp1.levels['data_version'], 'V3M2R1')

        stp2 = yeoda_path(self.test_dir, product='SSM', data_version=135, grid='EQUI7_EU500M',
                          tile='E048N012T6', qlook=True, make_dir=False)
        self.assertEqual(stp2.levels['data_version'], 'V135')

        with self.assertRaises(Exception) as context:
            stp3 = yeoda_path(self.test_dir, product='SSM', data_version='v12', grid='EQUI7_EU500M', tile='E048N012T6', qlook=True, make_dir=False)
        self.assertTrue(type(context.exception) == ValueError)


class TestYeodaTree(unittest.TestCase):
    """
    Tests checking if a yeoda tree is correctly reflected by yeoda_tree.
    """

    def setUp(self):
        """
        Setting up the test yeoda_tree.
        """

        self.test_dir = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), 'test_data', 'Sentinel-1_CSAR_IWGRDH')
        self.hierarchy_should = ['root', 'product', 'data_version', 'grid', 'tile', 'qlook']
        self.stt = yeoda_tree(self.test_dir, register_file_pattern='.tif')

    def test_tree_hierarchy(self):
        """
        Tests if a correct yeoda hierarchy was built.
        """

        self.assertEqual(self.stt.hierarchy, self.hierarchy_should)

        self.assertEqual(self.stt.root, self.test_dir)


    def test_tree_depth(self):
        """
        Checks if maximum depth is not violated.
        """

        max_depth_allowed = len(self.stt.root.split(os.sep)) + \
                            len(self.hierarchy_should) - 1

        self.assertTrue(all([len(x.split(os.sep)) <= max_depth_allowed
                             for x in self.stt.get_all_dirs()]))


class TestYeodaTreeSubset(unittest.TestCase):
    """
    Tests checking if a yeoda tree is correctly reflected by yeoda_tree.
    """

    def setUp(self):
        """
        Setting up the test yeoda_tree.
        """

        self.test_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_data', 'Sentinel-123')
        os.mkdir(self.test_dir)
        self.hierarchy_should = ['root', 'product', 'data_version', 'grid', 'tile', 'qlook']
        self.stt = yeoda_tree(self.test_dir, register_file_pattern='.tif')


    def tearDown(self):
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_subtree_subsetting(self):
        """
        Test the tree to handle subsetting correctly

        """

        stp1 = yeoda_path(self.test_dir, product='SSM', data_version='V3M2R1', grid='EQUI7_EU500M',
                          tile='E048N012T6', qlook=True, make_dir=True)
        stp2 = yeoda_path(self.test_dir, product='SSM', data_version='V3M2R1', grid='EQUI7_AF500M',
                          tile='E066N030T6', qlook=True, make_dir=True)
        stp3 = yeoda_path(self.test_dir, product='SSM', data_version='V3M2R1', grid='EQUI7_NA040M',
                          tile='E066N030T3', qlook=True, make_dir=True)

        os.makedirs(os.path.join(stp1.get_level('data_version'), 'logfiles', 'dummy'))

        # test get_subtree_matching() to get limited number of paths matching the level pattern.
        yt = yeoda_tree(self.test_dir, subset_pattern=('EQUI7', '500M'))
        self.assertEqual(yt.dir_count, 2)
        self.assertEqual(sorted(yt.collect_level_topnames('grid')), ['EQUI7_AF500M', 'EQUI7_EU500M'])

        # test get_subtree_unique_rebased() to get small, single, unique subtree,
        # which is is re-rooted to that level
        st = yt.get_subtree_unique_rebased('tile', 'E048N012T6')
        self.assertEqual(st.dir_count, 1)
        self.assertEqual(st.root, stp1.get_level('tile'))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from collections import OrderedDict

from geopathfinder.file_naming import SmartFilename


def decode_int(string):
    if isinstance(string, str):
        return int(string)
    else:
        return string


def encode_int(num):
    if isinstance(num, int):
        return str(num)
    else:
        return num


class TestSmartFilename(unittest.TestCase):

    def setUp(self):
        self.fields_def = OrderedDict([('pflag', {'len': 1}),
                                       ('dtime_1', {'len': 14})])

    def test_build_filename_wihout_ext(self):
        """
        Test building file naming without extension.
        """
        fields = {'pflag': 'M', 'dtime_1': '20180101120000'}
        smrtf = SmartFilename(fields, self.fields_def)

        self.assertEqual(str(smrtf), 'M_20180101120000')

    def test_build_filename_with_ext(self):
        """
        Test building file naming with extension.
        """
        fields = {'pflag': 'M', 'dtime_1': '20180101120000'}
        smrtf = SmartFilename(fields, self.fields_def, ext='.tif')

        self.assertEqual(str(smrtf), 'M_20180101120000.tif')

    def test_init_undefined_field(self):
        """
        Test initialization with undefined field name.
        """
        fields = {'pflag': 'M', 'test_time': '20180101120000'}

        with self.assertRaises(KeyError):
            SmartFilename(fields, self.fields_def, ext='.tif')

    def test_init_wrong_field_length(self):
        """
        Test initialization with wrong field length.
        """
        fields = {'pflag': 'ME'}

        with self.assertRaises(ValueError):
            SmartFilename(fields, self.fields_def, ext='.tif')

    def test_set_nonexisting_fields(self):
        """
        Test setting field which is non-existing in definition.
        """
        fields = {'pflag': 'M', 'dtime_1': '20180101120000'}
        smrtf = SmartFilename(fields, self.fields_def, ext='.tif')

        with self.assertRaises(KeyError):
            smrtf['new_field'] = 'test'

    def test_set_wrong_fields_len(self):
        """
        Test setting field with wrong length.
        """
        fields = {'pflag': 'M', 'dtime_1': '20180101120000'}
        smrtf = SmartFilename(fields, self.fields_def, ext='.tif')

        with self.assertRaises(ValueError):
            smrtf['pflag'] = 'MM'

    def test_set_and_get_fields(self):
        """
        Test set and get file name fields.
        """
        fields = {'pflag': 'M', 'dtime_1': '20180101120000'}
        smrtf = SmartFilename(fields, self.fields_def, ext='.tif')

        self.assertEqual(smrtf['pflag'], 'M')
        self.assertEqual(smrtf['dtime_1'], '20180101120000')

        smrtf['pflag'] = 'D'
        smrtf['dtime_1'] = '20180101130000'

        self.assertEqual(smrtf['pflag'], 'D')
        self.assertEqual(smrtf['dtime_1'], '20180101130000')

    def test_variable_length_fields(self):
        """
        Test if required delimiter for field with variable length is detected.
        """
        self.fields_def.update({'variable': {'len': 0, 'delim': ''}})

        fields = {'pflag': 'M', 'dtime_1': '20180101120000'}
        with self.assertRaises(ValueError):
            smrtf = SmartFilename(fields, self.fields_def)

        fn = 'M_20180101120000.tif'
        with self.assertRaises(ValueError):
            smrtf = SmartFilename.from_filename(fn, self.fields_def)

    def test_zero_entries(self):
        fields_def = OrderedDict([('text', {'len': 0}), ('empty', {'len': 0}), ('number', {'len': 6,
                                                                                           'decoder': decode_int,
                                                                                           'encoder': encode_int
                                                                                           })])
        fields = {'text': 'testfilename', 'number': 0}
        smrtf = SmartFilename(fields, fields_def, ext='.tif', compact=True)
        self.assertEqual(str(smrtf), 'testfilename__0.tif')
        smrtf_fn = SmartFilename.from_filename('fromfilename__0.tif', fields_def=fields_def, compact=True, convert=True)
        self.assertEqual(smrtf_fn['number'], 0)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import unittest
import os
import glob
import shutil

from geopathfinder.folder_naming import SmartPath
from geopathfinder.folder_naming import NullSmartPath
from geopathfinder.naming_conventions.sgrt_naming import sgrt_tree
from geopathfinder.folder_naming import transform_bytes

def cur_path():
    pth, _ = os.path.split(os.path.abspath(__file__))
    return pth


def get_test_sp(root,
                sensor=None,
                mode=None,
                group=None,
                datalog=None,
                product=None,
                wflow=None,
                grid=None,
                tile=None,
                var=None,
                qlook=True,
                make_dir=False):
    '''
    Function creating a SmartPath() for testing SmartPath().
    '''


    # defining the levels in the directory tree (order could become shuffled around)
    levels = {'root': root,
              'sensor': sensor,
              'mode': mode,
              'group': group,
              'datalog': datalog,
              'product': product,
              'wflow': wflow,
              'grid': grid,
              'tile': tile,
              'var': var,
              'qlook': 'qlooks'}

    # defining the hierarchy
    hierarchy = ['root', 'sensor', 'mode', 'group',
                 'datalog', 'product', 'wflow', 'grid',
                 'tile', 'var', 'qlook']

    return SmartPath(levels, hierarchy, make_dir=make_dir)


def get_test_sp_4_smarttree(sensor=None,
                mode=None,
                group=None,
                datalog=None,
                product=None,
                wflow=None,
                grid=None,
                tile=None,
                var=None,
                qlook=True,
                make_dir=False):
    '''
    Function creating a SmartPath() for testing SmartTree().
    '''


    # defining the levels in the directory tree (order could become shuffled around)
    levels = {'sensor': sensor,
              'mode': mode,
              'group': group,
              'datalog': datalog,
              'product': product,
              'wflow': wflow,
              'grid': grid,
              'tile': tile,
              'var': var,
              'qlook': 'qlooks'}

    # defining the hierarchy
    hierarchy = ['sensor', 'mode', 'group',
                 'datalog', 'product', 'wflow', 'grid',
                 'tile', 'var', 'qlook']

    return SmartPath(levels, hierarchy, make_dir=make_dir)


def disk_usage_filelist(filepaths, unit='KB'):
    bytes = sum([os.path.getsize(filepath) for filepath in filepaths])

    return transform_bytes(bytes, unit=unit)


class TestSmartPath(unittest.TestCase):

    def setUp(self):
        self.path = os.path.join(cur_path(), 'test_temp_dir')
        self.sp_obj = get_test_sp(self.path, sensor='Sentinel-1_CSAR',
                                  mode='IWGRDH', group='products',
                                  datalog='datasets', product='ssm',
                                  wflow='C1003', grid='EQUI7_EU500M',
                                  tile='E048N012T6', var='ssm',
                                  make_dir=True)

        self.path_perm = os.path.join(cur_path(), 'test_data')
        self.sp_obj_perm = get_test_sp(self.path_perm,
                                       sensor='Sentinel-1_CSAR',
                                       mode='IWGRDH', group='products',
                                       datalog='datasets', product='ssm',
                                       wflow='C1003', grid='EQUI7_EU500M',
                                       tile='E048N012T6', var='ssm',
                                       make_dir=True)

    def tearDown(self):
        if os.path.exists(self.path):
            shutil.rmtree(self.path)


    def test_get_dir(self):
        '''
        Testing the creation of the directory.

        '''
        result = self.sp_obj.get_dir(make_dir=True)

        assert os.path.exists(result)


    def test_build_levels(self):
        '''
        Testing the level creation.

        '''
        should = os.path.join(self.path, 'Sentinel-1_CSAR', 'IWGRDH', 'products',
                              'datasets', 'ssm', 'C1003')

        result = self.sp_obj.build_levels(level='wflow', make_dir=True)

        assert should == result

        assert os.path.exists(result)


    def test_get_level(self):
        '''
        Testing the level query.

        '''
        should = os.path.join(self.path, 'Sentinel-1_CSAR', 'IWGRDH')

        result = self.sp_obj['mode']

        assert should == result


    def test_expand_full_path(self):
        '''
        Testing the path expansion

        '''
        should = [os.path.join(self.path, 'Sentinel-1_CSAR', 'IWGRDH', 'MY_TEST.txt')]

        result = self.sp_obj.expand_full_path('mode', ['MY_TEST.txt'])

        assert should == result


    def test_search_files(self):
        '''
        Testing the file search yielding file lists.

        '''
        should = ['M20161218_051642--_SSM------_S1BIWGRDH1VVD_095_C1003_EU500M_E048N012T6.tif',
                  'M20170406_050911--_SSM------_S1AIWGRDH1VVD_022_C1003_EU500M_E048N012T6.tif']

        src = glob.glob(os.path.join(cur_path(), 'test_data', '*.*'))
        dest = self.sp_obj.build_levels(level='var', make_dir=True)

        for file in src:
            shutil.copy(file, dest)

        result = self.sp_obj.search_files('var', pattern='SSM')

        assert should == result


    def test_build_file_register(self):
        '''
        Testing the file register.

        '''
        self.sp_obj_perm.build_file_register(down_to_level='var')

        self.assertEqual(self.sp_obj_perm.file_count, 5)


    def test_get_disk_usage(self):
        '''
        Testing the disk usage function.

        '''
        pass
        du1 = self.sp_obj_perm.get_disk_usage('MB', up_to_level='grid')
        self.assertAlmostEqual(du1, 0.058, places=2)

        du2 = self.sp_obj_perm.get_disk_usage('MB', down_to_level='grid')
        self.assertAlmostEqual(du2, 1.032, places=2)

        du3 = self.sp_obj_perm.get_disk_usage('MB')
        self.assertAlmostEqual(du3, 1.090, places=2)

        self.assertAlmostEqual(du1 + du2, du3, places=2)


class TestSmartTree(unittest.TestCase):
    """
    Tests function of the SmartTree() class, applied for testing to the
    SGRT convention.

    """
    def setUp(self):
        self.test_dir = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), os.path.join('test_data', 'Sentinel-1_CSAR'))
        self.copy_dir = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), os.path.join('test_data', 'copy_target'))
        if os.path.exists(self.copy_dir):
            shutil.rmtree(self.copy_dir)
        os.mkdir(self.copy_dir)
        self.stt_1 = sgrt_tree(self.test_dir, register_file_pattern='.tif')


    def tearDown(self):
        if os.path.exists(self.copy_dir):
            shutil.rmtree(self.copy_dir)


    def test_tree_integrity(self):
        """
        Tests of only valid paths are added to the SmartTree()

        """
        self.assertTrue(
            all([True for x in self.stt_1.get_all_dirs() if self.test_dir in x]))


    def test_tree_dir_n_file_count(self):
        """
        Tests the dir_ and file_count of SmartTree()

        """
        self.assertEqual(self.stt_1.dir_count, 9)
        self.assertEqual(self.stt_1.file_count, 17)
        # TODO: should be 16 "file_too_deep.tif" should be not included in file_register!


    def test_get_smartpath(self):
        """
        Tests the selection of a SmartPath matching regex search patterns.

        """
        # typical use case
        should = ['M20160831_163321--_SIG0-----_S1AIWGRDH1VVA_175_A0201_EU500M_E048N006T6.tif']
        result = self.stt_1.get_smartpath(('A0202', 'sig0')).search_files(
            level='var')
        self.assertEqual(should, result)

        # typical use case
        should = ['Q20160831_163321--_SIG0-----_S1AIWGRDH1VVA_175_A0201_EU500M_E006N006T6.tif']
        result = self.stt_1[('A0202', 'sig0')].search_files(level='qlook')
        self.assertEqual(should, result)

        # test postive search pattern
        should = os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1003', 'EQUI7_EU500M', 'E048N012T6', 'ssm-noise', 'qlooks')
        result = self.stt_1.get_smartpath(('C1003', 'E048N012T6', 'noise')).get_dir()
        self.assertEqual(should, result)

        # test negative search pattern
        should = os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1003', 'EQUI7_EU500M', 'E048N012T6', 'ssm', 'qlooks')
        result = self.stt_1['C1003', 'E048N012T6', '-noise'].get_dir()
        self.assertEqual(should, result)

        # handling of no matches
        self.assertTrue(isinstance(self.stt_1['nonsense'], NullSmartPath))

        # handling of multiple matches
        self.assertTrue(isinstance(self.stt_1['A0202'], NullSmartPath))


    def test_collect_level(self):
        """
        Test the collect level functions

        """

        # unique case for the folder topnames
        should = ['E006N006T1', 'E006N006T6', 'E006N006T6',
                  'E006N006T6', 'E006N012T6', 'E048N012T6']
        result = sorted(self.stt_1.collect_level_topnames('tile', unique=True))
        self.assertEqual(should, result)

        # non-unique case for the folder topnames
        should = ['E006N006T1', 'E006N006T6', 'E006N006T6',
                  'E006N006T6', 'E006N006T6', 'E006N012T6',
                  'E048N012T6', 'E048N012T6']
        result = sorted(self.stt_1.collect_level_topnames('tile', unique=False))
        self.assertEqual(should, result)

        # unique case for full dirs
        should = [os.path.join(self.test_dir, 'IWGRDH', 'preprocessed', 'datasets', 'resampled', 'A0202', 'EQUI7_EU500M'),
                  os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1001', 'EQUI7_AF010M'),
                  os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1001', 'EQUI7_EU500M'),
                  os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1003', 'EQUI7_EU500M')]
        result = sorted(self.stt_1.collect_level_string('grid', unique=True))
        self.assertEqual(should, result)

        # test if unique full dirs deliver correct number of topnames
        should = ['EQUI7_AF010M',
                  'EQUI7_EU500M',
                  'EQUI7_EU500M',
                  'EQUI7_EU500M']
        result = sorted(self.stt_1.collect_level_topnames('grid', unique=True))
        self.assertEqual(should, result)


    def test_collect_level_smartpaths(self):
        """
        Test the collect level functions

        """

        # unique case for the folder topnames
        should = ['E006N006T1', 'E006N006T6', 'E006N006T6',
                  'E006N006T6', 'E006N012T6', 'E048N012T6']
        result = sorted(self.stt_1.collect_level_topnames('tile', unique=True))
        self.assertEqual(should, result)

        # non-unique case for the folder topnames
        should = ['E006N006T1', 'E006N006T6', 'E006N006T6',
                  'E006N006T6', 'E006N006T6', 'E006N012T6',
                  'E048N012T6', 'E048N012T6']
        result = sorted(self.stt_1.collect_level_topnames('tile', unique=False))
        self.assertEqual(should, result)

        # unique case for full dirs
        should = [os.path.join(self.test_dir, 'IWGRDH', 'preprocessed', 'datasets', 'resampled', 'A0202', 'EQUI7_EU500M'),
                  os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1001', 'EQUI7_AF010M'),
                  os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1001', 'EQUI7_EU500M'),
                  os.path.join(self.test_dir, 'IWGRDH', 'products', 'datasets', 'ssm', 'C1003', 'EQUI7_EU500M')]
        result = sorted(self.stt_1.collect_level_string('grid', unique=True))
        self.assertEqual(should, result)

        # test if unique full dirs deliver correct number of topnames
        should = ['EQUI7_AF010M',
                  'EQUI7_EU500M',
                  'EQUI7_EU500M',
                  'EQUI7_EU500M']
        result = sorted(self.stt_1.collect_level_topnames('grid', unique=True))
        self.assertEqual(should, result)


    def test_trim2branch(self):
        """
        Test the function for returning a branch (subtree).

        """

        branch1 = self.stt_1.get_subtree_unique_rebased('wflow', 'C1003', register_file_pattern=('.'))
        self.assertEqual(branch1.collect_level_topnames('root'), ['C1003'])

        self.assertEqual(branch1.dir_count, 4)
        self.assertEqual(len(branch1.file_register), 12)
        self.assertEqual(branch1.file_count, 12)

        should = ['E006N006T6', 'E006N012T6', 'E048N012T6']
        self.assertEqual(sorted(branch1.collect_level_topnames('tile')), should)

        # handling of multiple matches
        branch2 = self.stt_1.get_subtree_unique_rebased('grid', 'EQUI7_EU500M')
        self.assertEqual(branch2.dir_count, 0)
        self.assertEqual(len(branch2.file_register), 0)
        self.assertEqual(branch2.file_count, 0)


    def test_copy_smarttree_on_fs(self):
        """
        Tests if the copy functions works properly.

        """
        self.stt_1.copy_smarttree_on_fs(self.copy_dir)

        files = next(os.walk(self.copy_dir))[2]
        file_count = sum([len(files) for r, d, files in os.walk(self.copy_dir)])

        self.assertEqual(file_count, 24)


    def test_copy_smarttree_on_fs_level_pattern(self):
        """
        Tests if the copy functions works properly for given level and pattern.

        """
        self.stt_1.copy_smarttree_on_fs(self.copy_dir,
                                        level='wflow', level_pattern='A0202')

        files = next(os.walk(self.copy_dir))[2]
        file_count = sum([len(files) for r, d, files in os.walk(self.copy_dir)])

        self.assertEqual(file_count, 4)


    def test_get_disk_usage(self):
        """
        Tests if the disk usage functions works properly.

        """
        ssm_rel_filepaths = [
        ("IWGRDH", "products", "datasets", "ssm", "C1001", "EQUI7_AF010M", "E006N006T1", "ssm",
         "M20161017_053649--_SSM------_S1BIWGRDH1VVD_066_C1003_AF010M_E006N006T1.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1001", "EQUI7_EU500M", "E006N006T6", "ssm",
         "M20160426_053533--_SSM------_S1AIWGRDH1VVD_066_C1001_EU500M_E006N006T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1003", "EQUI7_EU500M", "E006N012T6", "ssm",
         "M20161005_053649--_SSM------_S1BIWGRDH1VVD_066_C1003_EU500M_E006N012T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1003", "EQUI7_EU500M", "E006N012T6", "ssm",
         "M20161017_053649--_SSM------_S1BIWGRDH1VVD_066_C1003_EU500M_E006N012T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1003", "EQUI7_EU500M", "E006N012T6", "ssm",
         "M20161029_053649--_SSM------_S1BIWGRDH1VVD_066_C1003_EU500M_E006N012T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm","C1003", "EQUI7_EU500M", "E048N012T6", "ssm",
         "M20150818_053449--_SSM------_S1AIWGRDH1VVD_066_C1003_EU500M_E048N012T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1003", "EQUI7_EU500M", "E048N012T6", "ssm",
         "M20170403_053310--_SSM------_S1BIWGRDH1VVD_066_C1003_EU500M_E048N012T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1003", "EQUI7_EU500M", "E048N012T6", "ssm", "qlooks",
         "Q20150818_053449--_SSM------_S1AIWGRDH1VVD_066_C1003_EU500M_E048N012T6.tif"),
        ("IWGRDH", "products", "datasets", "ssm", "C1003", "EQUI7_EU500M", "E048N012T6", "ssm", "qlooks",
         "Q20170403_053310--_SSM------_S1BIWGRDH1VVD_066_C1003_EU500M_E048N012T6.tif")]

        sig0_rel_filepaths = [
        ("IWGRDH", "preprocessed", "datasets", "resampled", "A0202", "EQUI7_EU500M", "E006N006T6",
         "sig0", "M20160831_163321--_SIG0-----_S1AIWGRDH1VVA_175_A0201_EU500M_E048N006T6.tif"),
        ("IWGRDH", "preprocessed", "datasets", "resampled", "A0202", "EQUI7_EU500M", "E006N006T6",
         "sig0", "qlooks", "Q20160831_163321--_SIG0-----_S1AIWGRDH1VVA_175_A0201_EU500M_E006N006T6.tif")]

        ssm_filepaths = [os.path.join(self.test_dir, *ssm_rel_filepath) for ssm_rel_filepath in ssm_rel_filepaths]
        sig0_filepaths = [os.path.join(self.test_dir, *sig0_rel_filepath) for sig0_rel_filepath in sig0_rel_filepaths]
        ssm_du = disk_usage_filelist(ssm_filepaths, unit='KB')
        sig0_du = disk_usage_filelist(sig0_filepaths, unit='KB')

        # test result from group_by
        result = self.stt_1.get_disk_usage(unit='KB', group_by=['var'])
        self.assertAlmostEqual(result.loc['ssm']['du'], ssm_du, places=2)

        # test result for file pattern
        result = self.stt_1.get_disk_usage(unit='KB', file_pattern='SIG0', total=True)
        self.assertAlmostEqual(result['du'][0], sig0_du, places=2)

        # test complete query result
        result = self.stt_1.get_disk_usage(unit='KB')
        self.assertEqual(result.shape, (33, 10))
        should = ['E006N006T1', 'E006N006T1', 'E006N006T6', 'E006N006T6', 'E006N006T6', 'E006N006T6', 'E006N006T6',
                  'E006N006T6', 'E006N006T6', 'E006N006T6', 'E006N006T6', 'E006N012T6', 'E006N012T6', 'E048N012T6',
                  'E048N012T6', 'E048N012T6', 'E048N012T6', 'E048N012T6']
        self.assertEqual(sorted(result['tile'].dropna().values), should)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
SGRT folder and file name definition.

"""

import os
import copy

import datetime as dt
from datetime import datetime
from collections import OrderedDict

from geopathfinder.folder_naming import build_smarttree
from geopathfinder.folder_naming import create_smartpath
from geopathfinder.file_naming import SmartFilename


# Please add here new sensors if they follow the SGRT naming convention.
allowed_sensor_dirs = ['Sentinel-1_CSAR',
                       'SCATSAR',
                       'METOP_ASCAT',
                       'Envisat_ASAR']


class SgrtFilename(SmartFilename):

    """
    SGRT file name definition using SmartFilename class.
    """

    fields_def = OrderedDict([
        ('pflag', {'len': 1, 'delim': ''}),
        ('dtime_1', {'len': 8}),
        ('dtime_2', {'len': 8}),
        ('var_name', {'len': 9}),
        ('mission_id', {'len': 2, 'delim': ''}),
        ('spacecraft_id', {'len': 1, 'delim': ''}),
        ('mode_id', {'len': 2, 'delim': ''}),
        ('product_type', {'len': 3, 'delim': ''}),
        ('res_class', {'len': 1, 'delim': ''}),
        ('level', {'len': 1, 'delim': ''}),
        ('pol', {'len': 2, 'delim': ''}),
        ('orbit_direction', {'len': 1}),
        ('relative_orbit', {'len': 3}),
        ('workflow_id', {'len': 5}),
        ('grid_name', {'len': 6}),
        ('tile_name', {'len': 10})
    ])
    pad = "-"
    delimiter = "_"

    def __init__(self, fields, ext=".tif", convert=False, compact=False):
        """
        Constructor of SgrtFilename class.

        Parameters
        ----------
        fields: dict
            Dictionary specifying the different parts of the filename.
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).
        """

        self.date_format = "%Y%m%d"
        self.time_format = "%H%M%S"
        fields = fields.copy()

        if 'dtime_2' not in fields.keys():
            self.single_date = True
            apply_dtime_2 = False
        else:
            self.single_date = False
            apply_dtime_2 = True

            if isinstance(fields['dtime_2'], str):
                if fields['dtime_2'].endswith('--'):
                    self.single_date = True
            else:
                if isinstance(fields['dtime_2'], dt.time) or (fields['dtime_2'].year < 1950):
                    self.single_date = True

        if 'dtime_1' in fields.keys():
            if self.single_date:
                date = self.encode_date(fields['dtime_1'])
                if apply_dtime_2:
                    time = self.encode_time(fields['dtime_2'])
                else:
                    time = self.encode_time(fields['dtime_1'])
                fields['dtime_1'] = date
                fields['dtime_2'] = time
            else:
                fields['dtime_1'] = self.encode_date(fields['dtime_1'])
                fields['dtime_2'] = self.encode_date(fields['dtime_2'])

        fields_def_ext = copy.deepcopy(SgrtFilename.fields_def)
        fields_def_ext['dtime_1']['decoder'] = lambda x: self.decode_date(x)
        fields_def_ext['dtime_1']['encoder'] = lambda x: self.encode_date(x)
        fields_def_ext['dtime_2']['decoder'] = lambda x: self.decode_time(x)
        fields_def_ext['dtime_2']['encoder'] = lambda x: self.encode_time(x)
        fields_def_ext['relative_orbit']['decoder'] = lambda x: self.decode_rel_orbit(x)
        fields_def_ext['relative_orbit']['encoder'] = lambda x: self.encode_rel_orbit(x)

        super(SgrtFilename, self).__init__(fields, fields_def_ext, ext=ext, pad=SgrtFilename.pad,
                                           delimiter=SgrtFilename.delimiter, convert=convert, compact=compact)

    @classmethod
    def from_filename(cls, filename_str, convert=False, compact=False):
        """
        Converts a filename given as a string into an SgrtFilename class object.

        Parameters
        ----------
        filename_str : str
            Filename without any paths (e.g., "M20170725_165004--_SIG0-----_S1BIWGRDH1VVA_146_A0104_EU500M_E048N012T6.tif").
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).

        Returns
        -------
        SgrtFilename
            Class representing an SGRT filename.
        """

        return super().from_filename(filename_str, SgrtFilename.fields_def, pad=SgrtFilename.pad,
                                     delimiter=SgrtFilename.delimiter, convert=convert, compact=compact)

    @property
    def stime(self):
        """
        Start time.

        Returns
        -------
        datetime.datetime
            Start time.
        """
        try:
            return datetime.combine(self["dtime_1"], self["dtime_2"]) if self.single_date else self["dtime_1"]
        except TypeError:
            return None

    @property
    def etime(self):
        """
        End time.

        Returns
        -------
        datetime.datetime
            End time.
        """
        try:
            return datetime.combine(self["dtime_1"], self["dtime_2"]) if self.single_date else self["dtime_2"]
        except TypeError:
            return None

    @property
    def time(self):
        """
        Unified time.

        Returns
        -------
        datetime.datetime
            Unified time.
        """
        try:
            if self.single_date:
                return self.stime
            else:
                return self.stime + (self.etime - self.stime) / 2
        except TypeError:
            return None

    @property
    def product_id(self):
        """
        Builds product id from other filename attributes (e.g. 'S1AIWGRDH').

        Returns
        -------
        product_id: str
            Product id consisting of mission id (e.g., 'S1'), spacecraft id (e.g., 'A'), mode id (e.g., 'IW'),
            product type (e.g., 'GRD') and resolution class (e.g., 'H').
        """
        try:
            product_id = "".join([self["mission_id"], self["spacecraft_id"], self["mode_id"], self["product_type"], self["res_class"]])
        except TypeError:
            product_id = None

        return product_id

    @property
    def ftile(self):
        """
        Builds the full tile name from other filename attributes (e.g. 'EU010M_E048N015T1').

        Returns
        -------
        ftile: str
            Full tile name consisting of grid name (e.g., 'EU10M') and tile name (e.g., 'E048N015T1').
        """
        try:
            ftile = "_".join([self["grid_name"], self["tile_name"]])
        except TypeError:
            ftile = None
        return ftile

    def decode_date(self, string):
        """
        Decodes a string into a datetime.date object. The format is given by the class.

        Parameters
        ----------
        string: str, object
            String needed to be decoded to a datetime.date object.

        Returns
        -------
        datetime.date, object
            Original object or datetime.date object parsed from the given string.
        """
        if isinstance(string, str):
            return datetime.strptime(string, self.date_format).date()
        else:
            return string

    def decode_time(self, string):
        """
        Decodes a string into a datetime.time/datetime.date object. The format is given by the class and the conversion
        follows the 'single_date' setting.

        Parameters
        ----------
        string: str, object
            String needed to be decoded to a datetime.time/datetime.date object.

        Returns
        -------
        datetime.date, datetime.time, object
            Original object, datetime.date or datetime.time object parsed from the given string.
        """
        if isinstance(string, str):
            if self.single_date:
                return datetime.time(datetime.strptime(string, self.time_format))
            else:
                return self.decode_date(string)
        else:
            return string

    def decode_rel_orbit(self, string):
        """
        Decodes a string into an integer.

        Parameters
        ----------
        string: str, object
            String needed to be decoded to an integer.

        Returns
        -------
        int, object
            Original object or integer object parsed from the given string.
        """
        if isinstance(string, str):
            return int(string)
        else:
            return string

    def encode_date(self, time_obj):
        """
        Encodes a datetime.datetime/datetime.date object into a string. The format is given by the class.

        Parameters
        ----------
        time_obj: datetime.datetime, datetime.date or object
            Datetime object needed to be encoded to a string.

        Returns
        -------
        str, object
            Original object or str object parsed from the given datetime object.
        """
        if isinstance(time_obj, (dt.datetime, dt.date, dt.time)):
            return time_obj.strftime(self.date_format)
        else:
            return time_obj

    def encode_time(self, time_obj):
        """
        Encodes a datetime.datetime/datetime.date object into a string. The format is given by the class.

        Parameters
        ----------
        time_obj: datetime.datetime, datetime.date or object
            Datetime object needed to be encoded to a string.

        Returns
        -------
        str, object
            Original object or str object parsed from the given datetime object.
        """
        if isinstance(time_obj, (dt.datetime, dt.time, dt.date)):
            if self.single_date:
                return time_obj.strftime(self.time_format)
            else:
                return time_obj.strftime(self.date_format)
        else:
            return time_obj

    def encode_rel_orbit(self, relative_orbit):
        """
        Encodes a relative orbit number into a string.

        Parameters
        ----------
        relative_orbit: int or object
            Integer needed to be encoded to a string.

        Returns
        -------
        str, object
            Original object or str object parsed from the given integer.
        """
        if isinstance(relative_orbit, int):
            return "{:03d}".format(relative_orbit)
        else:
            return relative_orbit


def sgrt_path(root, mode=None, group=None, datalog=None,
              product=None, wflow=None, grid=None, tile=None, var=None,
              qlook=True, make_dir=False):
    """
    Realisation of the full SGRT folder naming convention, yielding a single
    SmartPath.

    Parameters
    ----------
    root : str
        root directory of the path. must contain satellite sensor at toplevel.
        e.g. "R:\Datapool_processed\Sentinel-1_CSAR"
    mode : str
        e.g "IWGRDH"
    group : str, optional
        "preprocessed" or "parameters" or "products"
    datalog : str, optional
        must be "datasets" or "logfiles"
    product : str
        e.g. "ssm"
    wflow : str
        e.g. "C1003"
    grid : str
        e.g. "EQUI7_EU500M"
    tile : str
        e.g. "E048N012T6"
    var : str
        e.g. "ssm"
    qlook : bool
        if the quicklook subdir should be integrated
    make_dir : bool
        if the directory should be created on the filesystem

    Returns
    -------
    SmartPath
        Object for the path
    """

    # check the sensor folder name
    if root.split(os.sep)[-1] not in allowed_sensor_dirs:
        raise ValueError('Wrong input for "root"!')

    # define the datalog folder name
    if datalog is None:
        if isinstance(wflow, str):
            datalog = 'datasets'
    elif datalog == 'logfiles':
        product = None
        wflow = None
        grid = None
        tile = None
        var = None
        qlook = False
    elif datalog == 'datasets':
        pass
    else:
        raise ValueError('Wrong input for "datalog" level!')


    # define the group folder name
    if group is None:
        if wflow.startswith('A'):
            group = 'preprocessed'
        elif wflow.startswith('B'):
            group = 'parameters'
        elif wflow.startswith('C'):
            group = 'products'
        else:
            raise ValueError('Wrong input for "wflow" level!')


    # defining the folder levels
    levels = [mode, group,
              datalog, product, wflow, grid,
              tile,  var, 'qlooks']

    # defining the hierarchy
    hierarchy = ['mode', 'group',
                 'datalog', 'product', 'wflow', 'grid',
                 'tile', 'var', 'qlook']

    if qlook is False:
        levels.remove('qlooks')
        hierarchy.remove('qlook')

    return create_smartpath(root, hierarchy=hierarchy, levels=levels,
                     make_dir=make_dir)


def sgrt_tree(root, target_level=None, register_file_pattern=None,
              subset_level=None, subset_pattern=None, subset_unique=False):

    """
    Realisation of the full SGRT folder naming convention, yielding a
    SmartTree(), reflecting all subfolders as SmartPath()

    Parameters
    ----------
    root : str
        top level directory of the SGRT dataset, which is the sensor name in
        the SGRT naming convention.
        E.g.: "R:\Datapool_processed\Sentinel-1_CSAR"
    target_level : str, optional
        Can speed up things: Level name of target tree-depth.
        The SmartTree is only built from directories reaching this level,
        and only built down to this level. If not set, all directories are
        built down to deepest depth.
    register_file_pattern : str tuple, optional
        strings defining search pattern for file search for file_register
        e.g. ('C1003', 'E048N012T6').
        No asterisk is needed ('*')!
        Sequence of strings in given tuple is crucial!
        Be careful: If the tree is large, this can take a while!
    subset_level : str tuple, optional
        Name of level in tree's hierarchy where the subset should be applied
        e.g. ('tile').
        Default level is ('grid')
    subset_pattern : str tuple, optional
        Strings defining search pattern for subset_level, meaning only paths
        matching this pattern at "subset_level" will be included in the SmartTree().
        Default pattern is ('EQUI7').
        e.g. ('EQUI7', '500M'), or ('500M'). No asterisk is needed ('*')!
        Sequence of strings in given tuple is crucial!
    subset_unique : bool, optional
        defines of the subset will deliver...
            True: just one single subtree that matches uniquely the subset_pattern,
                  and which is rebased to the subset_level.
            False: all subtrees that match the subset_pattern (Default).

    Returns
    -------
    SmartTree
        Object for the SGRT tree.
    """

    # defining the hierarchy
    hierarchy = ['mode', 'group','datalog',
                 'product', 'wflow', 'grid',
                 'tile', 'var', 'qlook']

    # Check for allowed directory topnames for "root".
    if root.split(os.sep)[-1] in allowed_sensor_dirs:
        sgrt_tree = build_smarttree(root, hierarchy,
                                    target_level=target_level,
                                    register_file_pattern=register_file_pattern)
    else:
        raise ValueError('Root-directory "{}" does is '
                         'not a valid SGRT folder!'.format(root))

    # limit the tree to a subtree with all paths that match the subset_pattern at subset_level
    if subset_level is not None and not subset_unique:
        sgrt_tree = sgrt_tree.get_subtree_matching(subset_level, subset_pattern,
                                                   register_file_pattern=register_file_pattern)

    # limit the tree to a single, unique, small subtree that matches the subset_pattern at subset_level,
    # which is re-rooted to that level.
    elif subset_level is not None:
        sgrt_tree = sgrt_tree.get_subtree_unique_rebased(subset_level, subset_pattern,
                                                         register_file_pattern=register_file_pattern)

    return sgrt_tree


if __name__ == '__main__':
    pass



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import unittest
import logging
from datetime import datetime
from datetime import time

from geopathfinder.naming_conventions.acube_naming import ACubeFilename

logging.basicConfig(level=logging.INFO)


class TestACubeFilename(unittest.TestCase):

    def setUp(self):
        self.dtime_1 = datetime(2008, 1, 1, 12, 23, 33)
        self.dtime_2 = datetime(2009, 2, 2, 22, 1, 1)

        fields = {'dtime_1': self.dtime_1, 'dtime_2': self.dtime_2,
                  'var_name': 'SSM'}

        self.acube_fn = ACubeFilename(fields, convert=True)

        fields = {'dtime_1': self.dtime_1,
                  'var_name': 'SSM'}

        self.acube_fn2 = ACubeFilename(fields, convert=True)

        self.acube_fn3 = ACubeFilename(fields)

        fn = 'SIG0-----_SGRTA01_S1B_IWGRDH1VVA_20170725_165004--_EU500M_E048N012T6.tif'
        self.acube_fn4 = ACubeFilename.from_filename(fn, convert=True)

        fn = 'TMENSIG40_SGRTB01_ASA_WSM1-----D_20170725_20181225_EU500M_E048N012T6.tif'
        self.acube_fn5 = ACubeFilename.from_filename(fn)
        self.acube_fn6 = ACubeFilename.from_filename(fn, convert=True)

    def test1_build_acube_filename(self):
        """
        Test building ACube file name.

        """
        fn = ('SSM------_-------_---_----------_20080101_20090202_------_----------.tif')

        self.assertEqual(str(self.acube_fn), fn)

    def test2_get_n_set_date(self):
        """
        Test set and get start and end date.

        """
        self.assertEqual(self.acube_fn['dtime_1'], self.dtime_1.date())
        self.assertEqual(self.acube_fn['dtime_2'], self.dtime_2.date())

        new_start_time = datetime(2009, 1, 1, 12, 23, 33).date()
        self.acube_fn['dtime_1'] = new_start_time

        self.assertEqual(self.acube_fn['dtime_1'], new_start_time)

    def test30_get_n_set_date_n_time(self):
        """
        Test set and get date and time for a single datetime.

        """
        self.assertEqual(self.acube_fn2['dtime_1'], self.dtime_1.date())
        self.assertEqual(self.acube_fn2['dtime_2'], self.dtime_1.time())

        new_start_time = datetime(2009, 1, 1, 12, 23, 33)
        self.acube_fn2['dtime_1'] = new_start_time
        self.acube_fn2['dtime_2'] = new_start_time

        self.assertEqual(self.acube_fn2['dtime_1'], new_start_time.date())
        self.assertEqual(self.acube_fn2['dtime_2'], new_start_time.time())

    def test31_get_n_set_date_n_time_dts(self):
        """
        Test get and set date and time for a single datetime,
        returning strings.

        """
        self.assertEqual(self.acube_fn2.obj.dtime_1, datetime(2008, 1, 1).date())
        self.assertEqual(self.acube_fn2.obj.dtime_2, time(12, 23, 33))

        new_start_time = datetime(2345, 1, 2, 7, 8, 9)
        self.acube_fn2['dtime_1'] = new_start_time
        self.acube_fn2['dtime_2'] = new_start_time

        self.assertEqual(self.acube_fn2.obj.dtime_1, datetime(2345, 1, 2).date())
        self.assertEqual(self.acube_fn2.obj.dtime_2, time(7, 8, 9))

    def test32_get_n_set_date_n_time_strings(self):
        """
        Test get and set date and time for a single datetime,
        returning strings.

        """
        self.assertEqual(self.acube_fn3.obj.dtime_1, '20080101')
        self.assertEqual(self.acube_fn3.obj.dtime_2, '122333')

        new_start_time = datetime(2345, 1, 2, 7, 8, 9)
        self.acube_fn3['dtime_1'] = new_start_time
        self.acube_fn3['dtime_2'] = new_start_time

        self.assertEqual(self.acube_fn3.obj.dtime_1, '23450102')
        self.assertEqual(self.acube_fn3.obj.dtime_2, '070809')

    def test4_create_acube_filename(self):
        """
        Tests the creation of a SmartFilename from a given string filename.

        """

        # testing for single datetime
        self.assertEqual(self.acube_fn4['dtime_1'], datetime(2017, 7, 25).date())
        self.assertEqual(self.acube_fn4['dtime_2'], time(16, 50, 4))
        self.assertEqual(self.acube_fn4['var_name'], 'SIG0')
        self.assertEqual(self.acube_fn4['sat_name'], 'S1B')
        self.assertEqual(self.acube_fn4['product'], 'IWGRDH1')
        self.assertEqual(self.acube_fn4['pol'], 'VV')
        self.assertEqual(self.acube_fn4['direction'], 'A')
        self.assertEqual(self.acube_fn4['algorithm'], 'SGRTA01')
        self.assertEqual(self.acube_fn4['grid_name'], 'EU500M')
        self.assertEqual(self.acube_fn4['tile_name'], 'E048N012T6')
        self.assertEqual(self.acube_fn4.ext, '.tif')

        # testing for empty fields and two dates
        self.assertEqual(self.acube_fn5['dtime_1'], '20170725')
        self.assertEqual(self.acube_fn5['dtime_2'], '20181225')
        self.assertEqual(self.acube_fn5['var_name'], 'TMENSIG40')
        self.assertEqual(self.acube_fn5['sat_name'], 'ASA')
        self.assertEqual(self.acube_fn5['product'], 'WSM1')
        self.assertEqual(self.acube_fn5['pol'], '')

        # testing for empty pol field
        self.assertEqual(self.acube_fn6['pol'], None)

    def test5_build_ascat_ssm_fname(self):
        """
        Tests the creation of a resampled ASCAT SSM filename.

        """
        date_time = '20331122_112233'
        tilename = 'EU500M_E012N024T6'

        xfields = {'dtime_1': datetime.strptime(date_time[:8], "%Y%m%d"),
                   'dtime_2': datetime.strptime(date_time[-6:], "%H%M%S"),
                   'var_name': 'SSM',
                   'sat_name': 'ASC',
                   'product': 'SMO12NA',
                   'pol': 'XX',
                   'direction': 'D',
                   'algorithm': 'SGRTC01',
                   'grid_name': tilename[:6],
                   'tile_name': tilename[7:]}

        should = 'SSM------_SGRTC01_ASC_SMO12NAXXD_20331122_112233--_EU500M_E012N024T6.tif'
        fn = ACubeFilename(xfields)
        self.assertEqual(str(fn), should)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Module handling folder trees.
"""

import os
import glob
import copy
import shutil
import warnings

from datetime import datetime

import regex as re
import numpy as np
import pandas as pd


class SmartPath(object):

    """
    Base class for the single path structure to a data set.
    - allows building a path,
    - searching files with temporal slicing,
    - creating a pandas.DataFrame from a folder
    """

    def __init__(self, levels, hierarchy, make_dir=False):
        """

        Parameters
        ----------
        levels : dict
            dictionary assigning the name of levels to the hierarchy
        hierarchy : list of str
            List defining the order of the levels
        make_dir : bool, optional
            if set to True, then the full path of
            the SmartPath is created in the filesystem (default: False).
        """

        if all([x in hierarchy for x in list(levels.keys())]):

            self.levels = levels
            self.hierarchy = hierarchy

            directory = self.build_levels()
            self.directory = directory

            self.file_count = 0
            self.file_register = []
            self.has_register = False

            if make_dir:
                self.make_dir()

        else:
            raise ValueError('Levels are not reflected by the given hierarchy!')

    def __getitem__(self, level):
        """
        Short link for path, down to 'level'.
        Usage: path2level = your_smart_path[level]

        Parameters
        ----------
        level : str
            Name of level in hierarchy

        Returns
        -------
        path : str
            Path from root to level.
        """
        return self.get_level(level)


    def print_file_register(self):
        '''
        Nice function to print nicely all registered files to screen.
        '''

        print('\n'.join(self.file_register))


    def print_dir(self):
        '''
        Nice function to print nicely the directory of the path.
        '''

        print(self.directory)


    def get_dir(self, make_dir=False):
        """
        Get directory.

        Parameters
        ----------
        make_dir : bool, optional
            Create directory if not exists (default: False).

        Returns
        -------
        folder : str
            Full path of the SmartPath
        """
        if make_dir:
            self.make_dir()

        return self.directory

    def make_dir(self):
        """
        Creates directory from root to deepest level
        """
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)

    def build_levels(self, level='', make_dir=False):
        """

        Parameters
        ----------
        level : str, optional
            Name of level in hierarchy
        make_dir : bool, optional
            creates the directory until level

        Returns
        -------
        path : str
            Full path of the SmartPath (to the deepest level)
        """
        directory = r''

        for h in self.hierarchy:
            if self.levels[h] is None:
                break
            else:
                directory = os.path.join(directory, self.levels[h])
                if h == level:
                    break

        if make_dir:
            if not os.path.exists(directory):
                os.makedirs(directory)

        return directory

    def get_level(self, level):
        """
        Gets the path to the level.

        Parameters
        ----------
        level : str
            Name of level in hierarchy.

        Returns
        -------
        path : str
            Path from root to level.
        """
        if level in self.hierarchy:
            return self.build_levels(level=level)
        else:
            print('\'{}\' is not part of the path\'s hierarchy. '
                  'Try on of {}.'.format(level, self.hierarchy))


    def remove_level(self, level):
        '''
        In the SmartPath-instance, it removes a level from the hierarchy
        and level dictionary.

        Parameters
        ----------
        level : str
            Name of level in hierarchy.
        '''

        if level in self.hierarchy:

            self.levels.pop(level)
            self.hierarchy.remove(level)

            self.__init__(self.levels, self.hierarchy)

        else:
            print('Level \'{}\' is not in hierarchy!')


    def trim2level(self, level, remove='deeper_including'):
        '''
        Removes all levels that are higher or equal to given level.

        Parameters
        ----------
        level : str
            String of the level that should be removed, together will all
            higher levels.
        remove : str
            what should be removed?
            e.g. "deeper_including" removes the level itself, and deeper levels.
        '''

        dict = {'deeper_including': '>=',
                'deeper_excluding': '>',
                'higher_including': '<=',
                'higher_excluding': '<'}

        if level in self.hierarchy:

            hierarchy = copy.copy(self.hierarchy)
            level_ind = hierarchy.index(level)
            indxs = np.array(range(len(hierarchy)))
            cmd = 'np.array(hierarchy)[indxs {} {}].tolist()'.format(dict[remove], level_ind)
            subset = eval(cmd)

            for h in subset:
                self.remove_level(h)

        else:
            print('Level \'{}\' is not in hierarchy!')


    def base_onto_root(self, root):
        '''
        Adds as first level 'root' to the SmartPath-instance.

        Parameters
        ----------
        root : str
            String of the root directory
        '''

        if 'root' in self.hierarchy:
            self.remove_level('root')

        self.levels.update({'root': root})
        self.hierarchy = ['root'] + self.hierarchy

        self.__init__(self.levels, self.hierarchy)


    def expand_full_path(self, level, files):
        """
        Joins the path at level with given filenames.

        Parameters
        ----------
        level : str
            Name of level in hierarchy.
        files : list of str
            List of file names.

        Returns
        -------
        path : str
            Full file path.
        """
        return expand_full_path(self[level], files)


    def search_files(self, level, pattern=('.'), full_paths=False):
        """
        Searches files meeting the regex pattern at level in the SmartPath

        Parameters
        ----------
        level : str
            Name of level in hierarchy
        pattern : str tuple, optional
            strings defining search pattern for file search
            e.g. ('C1003', 'E048N012T6')
        full_paths : bool, optional
            If True, full paths will be included in dataframe (default: False)

        Returns
        -------
        filenames : list of str
            File names at the level.
        """
        if level not in self.levels.keys():
            return []
        else:
            return regex_file_search(self.build_levels(level), pattern,
                                 full_paths=full_paths)[0]


    def search_files_ts(self, level, pattern=('.'),
                        date_position=1, date_format='%Y%m%d_%H%M%S',
                        starttime=None, endtime=None, full_paths=False):
        """
        Function searching files at a level in the SmartPath,
        returning the filenames and the datetimes as pd.DataFrame

        Parameters
        ----------
        level : str
            name of level in hierarchy
        pattern : str tuple, optional
            strings defining search pattern for file search
            e.g. ('C1003', 'E048N012T6')
        date_position : int
            position of first character of date string in name of files
        date_format : str
            string with the datetime format in the filenames.
            e.g. '%Y%m%d_%H%M%S' reflects '20161224_000000'
        starttime : str or datetime, optional
            earliest date and time, if str must follow "date_format"
        endtime : str or datetime, optional
            latest date and time, if str must follow "date_format"
        full_paths : bool, optional
            should full paths be in the dataframe? default: False

        Returns
        -------
        df : pandas.DataFrame
            Dataframe holding the filenames and the datetimes
        """

        files = self.search_files(level, pattern=pattern)
        times = extract_times(
            files, date_position=date_position, date_format=date_format)

        if full_paths:
            files = self.expand_full_path(level, files)

        df = pd.DataFrame({'Files': files}, index=times)
        df.sort_index(inplace=True)

        if (starttime is not None) or (endtime is not None):
            if not isinstance(starttime, datetime):
                starttime = datetime.strptime(starttime, date_format)
            if not isinstance(endtime, datetime):
                endtime = datetime.strptime(endtime, date_format)
            df = df[starttime:endtime]

        return df


    def build_file_register(self, down_to_level=None, up_to_level=None, pattern=('.')):
        """
        Builds a file register collecting files at all levels in the SmartPath.

        Parameters
        ----------
        down_to_level : str, optional
            deepest level that should be included in the file register
        up_to_level : str, optional
            highest level that should be included in the file register
        pattern : str tuple, optional
            strings defining search pattern for file search
            e.g. ('C1003', 'E048N012T6')

        """
        file_register = []
        file_count = 0

        # limit the hierarchy
        if up_to_level is not None:
            idx = self.hierarchy[self.hierarchy.index(up_to_level):]
        elif down_to_level is not None:
            idx = self.hierarchy[:self.hierarchy.index(down_to_level) + 1]
        else:
            idx = self.hierarchy

        # search files at each level
        for h in idx:
            if self.levels[h] is not None:
                r, c = regex_file_search(self.build_levels(level=h), pattern, full_paths=True)
                file_register += r
                file_count += c

        self.file_register = file_register
        self.file_count = file_count
        self.has_register = True


    def get_disk_usage(self, unit=None, up_to_level=None, down_to_level=None, file_pattern=('.')):
        '''
        Computes the disk usage for each SmartPath and creates a Pandas DataFrame.

        Parameters
        ----------
        unit : str
            output unit of disk usage in bytes (e.g., "GB", "TB", ...)
        down_to_level : str, optional
            deepest level that should be included in the file register
        up_to_level : str, optional
            highest level that should be included in the file register
        file_pattern : str tuple, optional
            strings defining file pattern that are included in disk usage sums
            e.g. ('C1003', 'E048N012T6')

        Returns
        -------
        Number
            disk usage of all files along the SmartPath

        '''


        self.build_file_register(down_to_level=down_to_level, up_to_level=up_to_level,
                                 pattern=file_pattern)

        # compute size of directory
        nbytes = sum([os.path.getsize(filepath) for filepath in self.file_register])
        dir_size = transform_bytes(nbytes, unit=unit)

        return np.round(dir_size, 3)


class NullSmartPath(SmartPath):
    '''
    Class for non-exiting paths. Helps to avoid errors.

    '''
    def __init__(self):
        super(NullSmartPath, self).__init__({}, [], make_dir=False)


class SmartTree(object):
    '''
    Class for collecting multiple SmartPaths() at one root path.
    '''

    def __init__(self, root, hierarchy, make_dir=False):
        '''
        Initialises a SmartTree object for a given rootdirectory and hierarchy.

        Parameters
        ----------
        root : str
            directory path to root of directory tree.
        hierarchy : list of str
            List defining the order of the levels
        make_dir : bool, optional
            creates the root directory

        '''
        if not os.path.exists(root) and make_dir == False:
            raise OSError('Given directory for attribute \'root\', '
                          '\'{}\', does not exist!'.format(root))

        self.root = root
        self.dirs = {}
        self.hierarchy = hierarchy
        self.dir_count = 0
        self.file_count = 0
        self.file_register = []
        self.has_register = False

        if make_dir:
            if not os.path.exists(self.root):
                os.makedirs(self.root)


    def __getitem__(self, pattern):
        '''
        Shortcut for get_smartpath()

        Usage-example: smarttree['C1003', 'E048N012T6']
        '''

        return self.get_smartpath(pattern)


    def print_root(self):
        '''
        Function to print nicely the root to screen.

        '''

        print(self.root)


    def print_all_dirs(self):
        '''
        Nice function to print nicely all directories to screen.
        I'm proud of this function!
        '''

        print('\n'.join(self.get_all_dirs()))


    def print_file_register(self):
        '''
        Nice function to print nicely all registered files to screen.
        '''

        print('\n'.join(self.file_register))


    def print_collect_level(self, level, pattern=None, unique=False):
        '''
        Nice function to print nicely output from collect_level_string()
        '''

        print('\n'.join(self.collect_level_string(level, pattern=pattern,
                                                  unique=unique)))


    def print_collect_level_topnames(self, level, pattern=None, unique=True):
        '''
        Nice function to print nicely output from collect_level_topnames()
        '''

        print('\n'.join(self.collect_level_topnames(level, pattern=pattern,
                                                    unique=unique)))


    def file_register_search(self, pattern, full_paths=True):
        """
        Searches files in register meeting the regex pattern

        Parameters
        ----------
        pattern : str tuple, optional
            strings defining search pattern for file search
            e.g. ('C1003', 'E048N012T6')
        full_paths : bool, optional
            If True, full paths will be included in dataframe (default: False)
        Returns
        -------
        filenames : str or list of str
            File names

        """
        pattern = patterns_2_regex(pattern)

        regex = re.compile(pattern)
        files = [f for f in self.file_register if regex.match(os.path.basename(f))]

        if not full_paths:
            files = [os.path.basename(f) for f in files]

        if len(files) == 1:
            file = files[0]
            return file
        else:
            return sorted(files)


    def get_all_smartpaths(self):
        '''
        Returns SmartPaths in the SmartTree

        Returns
        -------
        list of SmartPaths
            List of all included SmartPaths
        '''

        return list(self.dirs.values())


    def get_all_dirs(self):
        '''
        Returns all full paths in the SmartTree

        Returns
        -------
        list
            Sorted list of all full paths
        '''

        return sorted(list(self.dirs.keys()))


    def get_disk_usage(self, unit='KB',
                       group_by=[],
                       file_pattern=('.'),
                       total=False):
        '''
        Computes the disk usage for each SmartPath and creates a Pandas DataFrame.

        Parameters
        ----------
        unit : str, optional
            output unit of disk usage in bytes (e.g., "GB", "TB", ...)
        group_by : list, optional
            list of levels forming groups, delivering disk usage sums
            e.g. ['tile', 'var']
        file_pattern : str tuple, optional
            strings defining file pattern that are included in disk usage sums
            e.g. ('M2019', 'SSM------')
        total : bool, optional
            returns the total disk usage for the root

        Returns
        -------
        DataFrame
            Pandas DataFrame containing the disk usage per SmartPath and the directory
            hierarchy as columns (without the root directory path)
        '''

        # hierarchy settings
        rootless_hierarchy = self.hierarchy[1:]
        n = len(rootless_hierarchy)

        # build SmartTree() disk usage table
        dir_size_table = []

        # loop over all levels of the hierarchy
        for i, level in enumerate(rootless_hierarchy):
            smpts_at_level = self.collect_level_smartpath(level, unique=True)

            # loop over trimmed paths
            for smpt in smpts_at_level:
                dir_elems = smpt.get_dir().replace(self.root, '').split(os.sep)
                dir_elems = map(lambda x: x.replace(os.sep, ''), dir_elems)
                dir_elems = [dir_elem for dir_elem in dir_elems if dir_elem != '']

                # fill up levels below with None
                remaining_levels = n - (i + 2) + 1
                dir_elems += [None] * remaining_levels

                filepaths = smpt.search_files(level, pattern=file_pattern, full_paths=True)
                filepaths = [filepath for filepath in filepaths if filepath in self.file_register]

                # get the disk usage of all files at the current level
                nbytes = sum([os.path.getsize(filepath) for filepath in filepaths])
                disk_usage = transform_bytes(nbytes, unit=unit)

                dir_elems.append(disk_usage)
                dir_size_table.append(dir_elems)

        # create Pandas DataFrame
        df = pd.DataFrame(data=dir_size_table, columns=self.hierarchy[1:] + ['du'])

        # return total disk usage
        if total:
            return pd.DataFrame(data=[[self.collect_level_topnames('root')[0], df['du'].sum()]], columns=['root', 'du'])
        # return disk usage of each SmartPaths
        elif group_by == []:
            return df
        # return disk usage summed over grouped SmartPaths
        else:
            return df.groupby(group_by).sum()


    def count_dirs(self):
        '''
        Sets the dir_count the SmartTree
        '''

        self.dir_count = len(self.dirs)


    def get_smartpath(self, pattern):
        '''
        Returns one SmartPath-object from the SmartTree that matches with
        the pattern. If more than one match, None is returned.

        Parameters
        ----------
        pattern : str tuple
            strings defining search pattern for path search
            e.g. ('C1003', 'E048N012T6')

        Returns
        -------
        SmartPath
            The path object matching the pattern.
        '''

        pattern = patterns_2_regex(pattern)

        paths = self.dirs.keys()

        matching_paths = []

        regex = re.compile(pattern)
        matching_paths += [m for m in paths if regex.match(m)]

        if len(matching_paths) == 0:
            warnings.warn('get_smartpath(): No matches for "pattern"!')
            return NullSmartPath()
        elif len(matching_paths) > 1:
            warnings.warn('get_smartpath(): Multiple matches for "pattern"!')
            return NullSmartPath()
        else:
            return self.dirs[matching_paths[0]]


    def collect_level_string(self, level, pattern=None, unique=False):
        '''
        Returns a list of paths at given level,
        and matching a given pattern.

        Parameters
        ----------
        level : str
            name of level in hierarchy
        pattern : str tuple, optional
            strings defining search pattern for path search
            e.g. ('C1003', 'E048N012T6')
        unique : bool, optional
            if set, a list of unique paths is returned
        Returns
        -------
        list of str
            list of paths at given level, matching the given pattern
        '''

        result = self.collect_level_smartpath(level=level, pattern=pattern, unique=unique)

        strings = [x.get_dir() for x in result]

        return strings


    def collect_level_smartpath(self, level, pattern=None, unique=False):
        '''
        Returns a list of Smartpaths reaching a given level,
        and matching a given pattern.

        Parameters
        ----------
        level : str
            name of level in hierarchy
        pattern : str tuple, optional
            strings defining search pattern for path search
            e.g. ('C1003', 'E048N012T6')
        unique : bool, optional
            if set, a list of unique paths is returned
        Returns
        -------
        list of SmartPaths
            list of paths at given level, matching the given pattern
        '''

        result = []

        for _, elem in self.dirs.items():
            smartpath = copy.deepcopy(elem)
            if smartpath.levels[level] is not None:
                if pattern is not None:
                    rpattern = patterns_2_regex(pattern)
                    regex = re.compile(rpattern)
                    if regex.match(smartpath.levels[level]):
                        smartpath.trim2level(level, remove='deeper_excluding')
                        result.append(smartpath)
                else:
                    smartpath.trim2level(level, remove='deeper_excluding')
                    result.append(smartpath)

        result = np.array(result)
        if unique:
            # remove duplicates after trimming the tree
            _, uni_idx = np.unique([x.get_dir() for x in result], return_index=True)
            result = result[uni_idx]

        sort_ind = np.argsort([x.get_dir() for x in result])

        return result[sort_ind].tolist()


    def collect_level_topnames(self, level, pattern=None, unique=True):
        '''
        Returns list of topnames of folders at given level.

        Parameters
        ----------
        level : str
            name of level in hierarchy
        pattern : str tuple, optional
            strings defining search pattern for path search
            e.g. ('C1003', 'E048N012T6')
        unique : bool, optional
            if set, a list of unique paths is returned

        Returns
        -------
        topnames : list of str
            list of folder-topnames at given level, matching the given pattern
        '''

        paths = self.collect_level_string(level, pattern=pattern, unique=unique)

        topnames = [x.split(os.sep)[-1] for x in paths]

        return topnames


    def add_smartpath(self, smartpath, make_dir=False):
        '''
        Adds a SmartPath-object to the SmartTree.

        Parameters
        ----------
        smartpath : SmartPath
            A SmartPath object. Only valid if hierarchy is compatible with the
            hierarchy of the SmartTree.
        make_dir : bool, optional
            creates the full directory of the SmartPath
        '''

        if isinstance(smartpath, SmartPath):

            smartpath.base_onto_root(self.root)

            if self.hierarchy == smartpath.hierarchy:

                self.dirs.update({smartpath.get_dir(): smartpath})
                self.count_dirs()

            else:
                print("SmartPath is not compatible with SmartTree: "
                      "Hierarchies do not correspond!")

            if make_dir:
                smartpath.make_dir()


    def remove_smartpath(self, key):
        '''
        Removes the SmartPath with 'key' from the SmartTree

        Parameters
        ----------
        key : str
            Path representing the key for the SmartPath
        '''

        self.dirs.pop(key)
        self.count_dirs()


    def get_subtree_matching(self, level, level_pattern, register_file_pattern=None):
        '''
        Returns a subtree of the SmartTree with branches comprising
        ALL matches with the pattern at the given level.

        Parameters
        ----------
        level : str
            Name of level in hierarchy.
            e.g. 'wflow'
        level_pattern : str
            string defining search pattern at given level
            e.g. 'C1003'
        register_file_pattern : str tuple, optional
            strings defining search pattern for file search for file_register
            e.g. ('C1003', 'E048N012T6')
            No asterisk is needed ('*')!
            Sequence of strings in given tuple is crucial!
            Be careful: If the tree is large, this can take a while!

        Returns
        -------
        branch : SmartTree
            SmartTree object that describes the seeked branch,
            part of current SmartTree
        '''

        branch = copy.deepcopy(self)

        branch_paths = self.collect_level_string(level, pattern=level_pattern, unique=True)

        if len(branch_paths) == 0:
            warnings.warn('get_subtree_unique_rebased(): No matches for "level_pattern" at "level"!')
            return NullSmartTree(self.root)
        else:
            for d in self.dirs.keys():
                match = [x for x in branch_paths if x in d]
                if len(match) == 0:
                    branch.remove_smartpath(d)

            # update stats
            branch.count_dirs()
            file_register = []
            file_count = 0
            # update file_register
            if register_file_pattern is not None:
                for sp in branch.dirs.values():
                    sp.build_file_register(pattern=register_file_pattern)
                    file_register += sp.file_register
                    file_count += sp.file_count

            branch.file_register = file_register
            branch.file_count = file_count
            branch.has_register = True

            return branch


    def get_subtree_unique_rebased(self, level, level_pattern, register_file_pattern=None):
        '''
        Returns a single branch (a subtree) of the SmartTree
        with ONE UNIQUE match with the pattern at the given level.
        Performs rebasing to deeper root.

        Parameters
        ----------
        level : str
            Name of level in hierarchy.
            e.g. 'wflow'
        level_pattern : str
            string defining search pattern at given level
            e.g. 'C1003'
        register_file_pattern : str tuple, optional
            strings defining search pattern for file search for file_register
            e.g. ('C1003', 'E048N012T6')
            No asterisk is needed ('*')!
            Sequence of strings in given tuple is crucial!
            Be careful: If the tree is large, this can take a while!

        Returns
        -------
        branch : SmartTree
            SmartTree object that describes the seeked branch,
            part of current SmartTree
        '''

        branch = copy.deepcopy(self)

        branch_path = self.collect_level_string(level, pattern=level_pattern, unique=True)

        if len(branch_path) == 0:
            warnings.warn('get_subtree_unique_rebased(): No matches for "level_pattern" at "level"!')
            return NullSmartTree(self.root)
        elif len(branch_path) > 1:
            warnings.warn('get_subtree_unique_rebased(): Multiple matches for "level_pattern" at "level"!')
            return NullSmartTree(self.root)
        else:
            for d in self.dirs.keys():
                if not branch_path[0] in d:
                    branch.remove_smartpath(d)
                else:
                    branch.dirs.get(d).trim2level(level, remove="higher_including")
                    branch.dirs.get(d).base_onto_root(branch_path[0])

            # update dir_count
            branch.count_dirs()
            # update tree root
            branch.root = branch_path[0]
            # update tree hierarchy
            for h in copy.copy(branch.hierarchy):
                if h != level:
                    branch.hierarchy.remove(h)
                else:
                    branch.hierarchy.remove(h)
                    break
            branch.hierarchy = ['root'] + branch.hierarchy

            file_register = []
            file_count = 0
            # update file_register
            if register_file_pattern is not None:
                for sp in branch.dirs.values():
                    sp.build_file_register(pattern=register_file_pattern)
                    file_register += sp.file_register
                    file_count += sp.file_count

            branch.file_register = file_register
            branch.file_count = file_count
            branch.has_register = True

            return branch

    # BBM: a function like this should be implemented, reducing all registered files and smartpaths to a filelist
    # def trim2filelist(self, file_list):
    #
    #     included_files = list(set(self.file_register) & set(file_list))
    #     file_count = len(included_files)
    #     if file_count != 0:
    #
    #         tree = copy.deepcopy(self)
    #         tree.file_register = included_files
    #
    #         tree2 = SmartTree(self.root, self.hierarchy)
    #         tree2.file_register = included_files
    #         tree2.file_count = file_count
    #
    #
    #     pass

    def make_dirs(self):
        '''
        Creates a full path for each of the contained SmartPaths.
        '''

        for _, smartpath in self.dirs.items():
            smartpath.make_dir()


    def copy_smarttree_on_fs(self, target_dir, level=None, level_pattern='',
                             file_pattern=None, file_list=None):
        """
        Copies all files and directories in the SmartTree to a target
        directory, using shutil.copytree().

        Parameters
        ----------
        target_dir : str
            the target directory
        level : str, optional
            if set, only the branch at given "level" and matching given
            "pattern" will be copied. e.g. 'wflow'.
            Otherwise all below root directory will be copied.
        level_pattern : str
            string defining search pattern at given level
            e.g. 'A0202'
        file_pattern : str or tuple of str, optional
            string patterns for file matching.
            when starting with "-" it is interpreted as negative pattern
            that should be excluded from the matches
        file_list : list of str, optional
            List with full file paths that should be included in the copy process.
            Only files that are in both: tree.file_register & file_list are copied!
            This list might come from a yeoda datacube filtering

        Returns
        -------

        """

        if level is None:
            source_dir = self.root
            base_folder = self.collect_level_topnames('root')[0]
        else:
            branch = self.get_subtree_unique_rebased(level, level_pattern, register_file_pattern=file_pattern)
            source_dir = branch.collect_level_string('root', level_pattern, unique=True)[0]
            base_folder = branch.collect_level_topnames('root')[0]

        target_dir = os.path.join(target_dir, base_folder)

        if file_list is None:
            copy_tree(source_dir, target_dir, file_pattern=file_pattern)
        else:
            included_files = list(set(self.file_register) & set(file_list))
            file_count = len(included_files)
            if file_count != 0:
                copy_tree(source_dir, target_dir, file_list=included_files)


class NullSmartTree(SmartTree):
    '''
    Class for non-exiting paths. Helps to avoid errors.
    '''
    def __init__(self, root):
        super(NullSmartTree, self).__init__(root, [], make_dir=False)


def create_smartpath(root, hierarchy, levels, make_dir=False):
    '''
    Function for creating a SmartPath().

    Parameters
    ----------
    root : str
        root path of the SmartPath. Gets added as level 'root' in hierarchy.
    hierarchy : list of str
        list defining the order of the levels
    levels : list
        list of the names of levels in the hierarchy
    make_dir : bool, optional
        if set to True, then the full path of
        the SmartPath is created in the filesystem (default: False).

    Returns
    -------
    SmartPath

    '''

    hierarchy = ['root'] + hierarchy
    levels = [root] + levels

    # defining the levels in the directory
    xlevels = {}
    for p in range(len(hierarchy)):
        xlevels.update({hierarchy[p]: levels[p]})

    return SmartPath(xlevels, hierarchy, make_dir=make_dir)


def build_smarttree(root,
                    hierarchy,
                    target_level=None,
                    register_file_pattern=None):
    '''
    Function walking through directories in root path for building a structure
    of SmartPaths. Can also search for files.
    Attention: The SmartTree is only working properly if all folders in "root"
    follow the "hierarchy"!

    Parameters
    ----------
    root : str
        root path of the SmartTree. Gets added as level 'root' in hierarchy.
    hierarchy : list of str
        List defining the order of the levels
    target_level : str, optional
        Can speed up things: Level name of target tree-depth.
        The SmartTree is only built from directories reaching this level,
        and only built down to this level. If not set, all directories are
        built down to deepest depth.
    register_file_pattern : str tuple, optional
        strings defining search pattern for file search for file_register
        e.g. ('C1003', 'E048N012T6')
        No asterisk is needed ('*')!
        Sequence of strings in given tuple is crucial!
        Be careful: If the tree is large, this can take a while!

    Returns
    -------
    SmartTree
        Tree object for the dataset.

    '''

    # initialises the SmartTree
    smart_tree = SmartTree(root, ['root'] + hierarchy, make_dir=False)

    # path depth of root
    root_depth = len(root.split(os.sep))

    alldirs = []
    depth = []

    # walk thru the dirs below of root
    for dirpath, dirs, files in os.walk(root, topdown=False):
        alldirs += [dirpath.replace(root, '')]
        depth += [len(dirpath.split(os.sep)) - root_depth]
        # if set, then files are registered
        # (they are in the memory anyway at this moment)
        if register_file_pattern is not None:
            files, count = regex_file_search(dirpath,
                                             register_file_pattern,
                                             full_paths=True)
            smart_tree.file_register += files
            smart_tree.file_count += count
    alldirs = np.array(alldirs)

    # only select paths reaching given target level
    if target_level is not None:
        target_depth = smart_tree.hierarchy.index(target_level)
        singular_paths = alldirs[np.array(depth) == target_depth]

        # reset file register (to register only files down to target level)
        file_register = []

    # select all singular paths that have no children - and drop their parents
    else:
        target_depth = max(depth)
        singular_paths = alldirs[np.array(depth) == target_depth]
        for d in sorted((set(depth)), reverse=True):
            if d == target_depth:
                continue
            paths_at_depth = alldirs[np.array(depth) == d]
            paths_ending = []
            for pad in paths_at_depth:
                if all([pad not in x for x in singular_paths]):
                    paths_ending += [pad]

            singular_paths = np.append(singular_paths, paths_ending)

    # create and append a SmartPath for each singular path
    for fp in singular_paths:
        levels = {}
        sub_levels = fp.split(os.sep)[1:]
        tail_depth = len(sub_levels)
        for p in range(len(hierarchy)):
            if p < tail_depth:
                levels.update({hierarchy[p]: sub_levels[p]})
            else:
                # note sure about this. Potentially causes problems somewhere
                levels.update({hierarchy[p]: None})

        smart_path = SmartPath(levels, hierarchy)

        smart_tree.add_smartpath(smart_path)

        # to register only files down to target level
        if register_file_pattern is not None and target_level is not None:
            smart_path.build_file_register(down_to_level=target_level,
                                           pattern=register_file_pattern)
            file_register += smart_path.file_register

        smart_path = None
        levels = None

    # update self.dir_count
    smart_tree.count_dirs()

    # register only files in paths down to target level
    if register_file_pattern is not None and target_level is not None:
        smart_tree.file_register = list(set(file_register))
        smart_tree.file_count = len(smart_tree.file_register)

    if register_file_pattern is not None:
        smart_tree.has_register = True

    return smart_tree


def expand_full_path(path, files):
    """
    Joins the path at level with given filenames.

    Parameters
    ----------
    level : str
        Name of level in hierarchy.
    files : list of str
        List of file names.

    Returns
    -------
    path : str
        Full file path.
    """
    return [os.path.join(path, f) for f in files]


def reduce_2_basename(files):
    """
    Converts full file paths to file base names.

    Parameters
    ----------
    files : list of str
        list of filepaths

    Returns
    -------
    filenames : list of str
        List of base file names.
    """
    return [os.path.basename(f) for f in files]


def extract_times(files, date_position=1, date_format='%Y%m%d_%H%M%S'):
    """
    Extracts the datetimes from filenames.

    Parameters
    ----------
    files : list of str
        list of strings with filenames or filepaths
    date_position : int
        position of first character of date string in name of files
    date_format: str
        string with the datetime format in the filenames.
        '%Y%m%d_%H%M%S' reflects eg. '20161224_000000'

    Returns
    -------
    times : list of datetime
        List of datetime objects extracted from the filenames.
    """
    if any([os.path.isdir(x) for x in files]):
        files = reduce_2_basename(files)

    times = []
    for f in files:
        t = datetime.strptime(
            f[date_position:date_position + len(date_format) + 2],
            date_format)
        times.append(t)

    return times


def patterns_2_regex(patterns):
    '''
    Converts any string, or tuple of strings, to a regex pattern.

    Parameters
    ----------
    patterns : str or tuple of str
        string patterns for matching.
        when starting with "-" it is interpreted as negative pattern
        that should be excluded from the matches

    Returns
    -------
    str: regex string
    '''

    if isinstance(patterns, str):
        patterns = [patterns]

    regex = [''] * len(patterns)

    for n, p in enumerate(patterns):

        # negative pattern (leads to exclusion)
        if p.startswith('-'):
            regex[n] = '((?!{}).)*$'.format(p[1:])

        # positive pattern (leads to inclusion)
        else:
            regex[n] = '.*{}'.format(p)

    return ''.join(regex)


def regex_file_search(path, pattern, full_paths=True):
    '''
    Carries out the file search using the strings in pattern as regex strings.

    Parameters
    ----------
    path : search in this directory. Subdirectories are ignored.
    pattern : str or tuple of str
        string patterns for matching.
        when starting with "-" it is interpreted as negative pattern
        that should be excluded from the matches
    full_paths : bool, optional
        should full paths be returned? default: True

    Returns
    -------
    tuple
        a tuple (files, count) that contains the file list and the
        count of files

    '''

    pattern = patterns_2_regex(pattern)

    paths = glob.glob(os.path.join(path, '*.*'))
    basenames = reduce_2_basename(paths)

    regex = re.compile(pattern)
    files = [f for f in basenames if regex.match(f)]

    if full_paths:
        files = expand_full_path(path, files)

    return sorted(files), len(files)


def copy_tree(source, dest, file_pattern=None, overwrite=False,
              file_list=None):
    """
    Copies a directory tree structure.

    Parameters
    ----------
    source : str
        directory that should be copied, recursively
    dest : str
        where the tree should be copied to
    file_pattern : str or tuple of str, optional
        string patterns for file matching.
        when starting with "-" it is interpreted as negative pattern
        that should be excluded from the matches
    overwrite : bool, optional
        should existing files be overwritten? default: False
    file_list : list of str, optional
        List with full file paths that should be included in the copy process.
        This list might come from a yeoda datacube filtering

    """

    # only files matching regex pattern are copied
    if file_pattern is not None:
        fpattern = patterns_2_regex(file_pattern)
        regex = re.compile(fpattern)

    for root, dirs, files in os.walk(source):
        if not os.path.isdir(root):
            os.makedirs(root)

        # all files found
        for file in files:

            if file_list is not None:
                if os.path.join(root, file) not in file_list:
                    continue

            if file_pattern is not None:
                include = regex.match(file)
            else:
                include = True

            # proceed if match is given or not required
            if include:
                rel_path = root.replace(source, '').lstrip(os.sep)
                dest_path = os.path.join(dest, rel_path)

                # clarify status for over(writing)
                file2write = os.path.join(dest_path, file)
                file_exists = (os.path.exists(file2write))
                do_writing = file_exists and overwrite or not file_exists
                if do_writing:

                    if not os.path.isdir(dest_path):
                        os.makedirs(dest_path)

                    # do it. with file metadata (copy2)
                    shutil.copy2(os.path.join(root, file), file2write)


def transform_bytes(bytes, unit='KB'):
    scale_factor_dict = {"TB": 1e-12, "GB": 1e-9, "MB": 1e-6, "KB": 1e-3}
    scale_factor = 1.
    if unit is not None:
        try:
            scale_factor = scale_factor_dict[unit.upper()]
        except KeyError:
            raise KeyError('Unit {} unknown.'.format(unit))

    return bytes*scale_factor

if __name__ == '__main__':
    pass


# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
ACube file name definition.

"""

import copy

import datetime as dt
from datetime import datetime
from collections import OrderedDict
from geopathfinder.file_naming import SmartFilename


class ACubeFilename(SmartFilename):

    """
    ACube file name definition using SmartFilename class.
    """

    fields_def = OrderedDict([
        ('var_name', {'len': 9}),
        ('algorithm', {'len': 7}),
        ('sat_name', {'len': 3}),
        ('product', {'len': 7, 'delim': ''}),
        ('pol', {'len': 2, 'delim': ''}),
        ('direction', {'len': 1}),
        ('dtime_1', {'len': 8}),
        ('dtime_2', {'len': 8}),
        ('grid_name', {'len': 6}),
        ('tile_name', {'len': 10})])
    pad = "-"
    delimiter = "_"

    def __init__(self, fields, ext=".tif", convert=False, compact=False):
        """
        Constructor of ACube Filename class.

        Parameters
        ----------
        fields: dict
            Dictionary specifying the different parts of the filename.
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).
        """

        self.date_format = "%Y%m%d"
        self.time_format = "%H%M%S"
        fields = fields.copy()

        if 'dtime_2' not in fields.keys():
            self.single_date = True
            apply_dtime_2 = False
        else:
            self.single_date = False
            apply_dtime_2 = True

            if isinstance(fields['dtime_2'], str):
                if fields['dtime_2'].endswith('--'):
                    self.single_date = True
            else:
                if isinstance(fields['dtime_2'], dt.time) or (fields['dtime_2'].year < 1950):
                    self.single_date = True

        if 'dtime_1' in fields.keys():
            if self.single_date:
                date = self.encode_date(fields['dtime_1'])
                if apply_dtime_2:
                    time = self.encode_time(fields['dtime_2'])
                else:
                    time = self.encode_time(fields['dtime_1'])
                fields['dtime_1'] = date
                fields['dtime_2'] = time
            else:
                fields['dtime_1'] = self.encode_date(fields['dtime_1'])
                fields['dtime_2'] = self.encode_date(fields['dtime_2'])

        fields_def_ext = copy.deepcopy(ACubeFilename.fields_def)
        fields_def_ext['dtime_1']['decoder'] = lambda x: self.decode_date(x)
        fields_def_ext['dtime_1']['encoder'] = lambda x: self.encode_date(x)
        fields_def_ext['dtime_2']['decoder'] = lambda x: self.decode_time(x)
        fields_def_ext['dtime_2']['encoder'] = lambda x: self.encode_time(x)

        super(ACubeFilename, self).__init__(fields, fields_def_ext, ext=ext, pad=ACubeFilename.pad,
                                            delimiter=ACubeFilename.delimiter, convert=convert, compact=compact)

    @classmethod
    def from_filename(cls, filename_str, convert=False, compact=False):
        """
        Converts a filename given as a string into an ACubeFilename class object.

        Parameters
        ----------
        filename_str : str
            Filename without any paths (e.g., "SIG0-----_SGRTA01_S1A_IWGRDH1VHD_20160126_054245--_EU010M_E043N016T1.tif").
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).

        Returns
        -------
        ACubeFilename
            Class representing an ACube file name.
        """

        return super().from_filename(filename_str, ACubeFilename.fields_def, pad=ACubeFilename.pad,
                                     delimiter=ACubeFilename.delimiter, convert=convert, compact=compact)

    @property
    def stime(self):
        """
        Start time.

        Returns
        -------
        datetime.datetime
            Start time.
        """
        try:
            return datetime.combine(self["dtime_1"], self["dtime_2"]) if self.single_date else self["dtime_1"]
        except TypeError:
            return None

    @property
    def etime(self):
        """
        End time.

        Returns
        -------
        datetime.datetime
            End time.
        """
        try:
            return datetime.combine(self["dtime_1"], self["dtime_2"]) if self.single_date else self["dtime_2"]
        except TypeError:
            return None

    @property
    def time(self):
        """
        Unified time.

        Returns
        -------
        datetime.datetime
            Unified time.
        """
        try:
            if self.single_date:
                return self.stime
            else:
                return self.stime + (self.etime - self.stime) / 2
        except TypeError:
            return None

    @property
    def ftile(self):
        """
        Builds the full tile name from other filename attributes (e.g. 'EU010M_E048N015T1').

        Returns
        -------
        ftile: str
            Full tile name consisting of grid name (e.g., 'EU10M') and tile name (e.g., 'E048N015T1').
        """
        try:
            ftile = "_".join([self["grid_name"], self["tile_name"]])
        except TypeError:
            ftile = None
        return ftile

    def decode_date(self, string):
        """
        Decodes a string into a datetime.date object. The format is given by the class.

        Parameters
        ----------
        string: str, object
            String needed to be decoded to a datetime.date object.

        Returns
        -------
        datetime.date, object
            Original object or datetime.date object parsed from the given string.
        """
        if isinstance(string, str):
            return datetime.strptime(string, self.date_format).date()
        else:
            return string

    def decode_time(self, string):
        """
        Decodes a string into a datetime.time/datetime.date object. The format is given by the class and the conversion
        follows the 'single_date' setting.

        Parameters
        ----------
        string: str, object
            String needed to be decoded to a datetime.time/datetime.date object.

        Returns
        -------
        datetime.date, datetime.time, object
            Original object, datetime.date or datetime.time object parsed from the given string.
        """
        if isinstance(string, str):
            if self.single_date:
                return datetime.time(datetime.strptime(string, self.time_format))
            else:
                return self.decode_date(string)
        else:
            return string

    def encode_date(self, time_obj):
        """
        Encodes a datetime.datetime/datetime.date object into a string. The format is given by the class.

        Parameters
        ----------
        time_obj: datetime.datetime, datetime.date or object
            Datetime object needed to be encoded to a string.

        Returns
        -------
        str, object
            Original object or str object parsed from the given datetime object.
        """
        if isinstance(time_obj, (dt.datetime, dt.date, dt.time)):
            return time_obj.strftime(self.date_format)
        else:
            return time_obj

    def encode_time(self, time_obj):
        """
        Encodes a datetime.datetime/datetime.date object into a string. The format is given by the class.

        Parameters
        ----------
        time_obj: datetime.datetime, datetime.date or object
            Datetime object needed to be encoded to a string.

        Returns
        -------
        str, object
            Original object or str object parsed from the given datetime object.
        """
        if isinstance(time_obj, (dt.datetime, dt.time, dt.date)):
            if self.single_date:
                return time_obj.strftime(self.time_format)
            else:
                return time_obj.strftime(self.date_format)
        else:
            return time_obj

if __name__ == '__main__':
    pass



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
BMon file name definition.

"""

import copy
import datetime as dt
from datetime import datetime
from collections import OrderedDict

from geopathfinder.file_naming import SmartFilename


class BMonFilename(SmartFilename):

    """
    BMon file name definition using SmartFilename class.
    """

    fields_def = OrderedDict([
        ('var_name', {'len': 16}),
        ('sres', {'len': 4, 'start': 17}),
        ('timestamp', {'len': 14}),
        ('version', {'len': 2})
    ])
    pad = "-"
    delimiter = "_"

    def __init__(self, fields, ext='.nc', convert=False, compact=False):
        """
        Constructor of BMonFilename class.

        Parameters
        ----------
        fields: dict
            Dictionary specifying the different parts of the filename.
        ext : str, optional
            File name extension (default is '.nc').
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).
        """

        self.timestamp_format = "%Y%m%d%H%M%S"

        fields_def_ext = copy.deepcopy(BMonFilename.fields_def)
        fields_def_ext['timestamp']['decoder'] = lambda x: self.decode_timestamp(x)
        fields_def_ext['timestamp']['encoder'] = lambda x: self.encode_timestamp(x)

        super(BMonFilename, self).__init__(fields, fields_def_ext, ext=ext, pad=BMonFilename.pad,
                                           delimiter=BMonFilename.delimiter, convert=convert, compact=compact)

    @classmethod
    def from_filename(cls, filename_str, convert=False, compact=False):
        """
        Converts a filename given as a string into a BMonFilename class object.

        Parameters
        ----------
        filename_str : str
            Filename without any paths (e.g., "BMON_DM_ENSEMBLE_500m_20160101120000_v1.nc").
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).

        Returns
        -------
        BMonFilename
            Class representing a BMON filename.
        """

        return super().from_filename(filename_str, BMonFilename.fields_def, pad=BMonFilename.pad,
                                     delimiter=BMonFilename.delimiter, convert=convert, compact=compact)

    def decode_timestamp(self, string):
        """
        Decodes a string into a datetime.date object. The format is given by the class.

        Parameters
        ----------
        string: str, object
            String needed to be decoded to a datetime.date object.

        Returns
        -------
        datetime.date, object
            Original object or datetime.date object parsed from the given string.
        """

        if isinstance(string, str):
            return datetime.strptime(string, self.timestamp_format)
        else:
            return string

    def encode_timestamp(self, time_obj):
        """
        Encodes a datetime.datetime/datetime.date object into a string. The format is given by the class.

        Parameters
        ----------
        time_obj: datetime.datetime, datetime.date or object
            Datetime object needed to be encoded to a string.

        Returns
        -------
        str, object
            Original object or str object parsed from the given datetime object.
        """

        if isinstance(time_obj, (dt.datetime, dt.date, dt.time)):
            return time_obj.strftime(self.timestamp_format)
        else:
            return time_obj

if __name__ == '__main__':
    pass



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import unittest
from datetime import datetime

from geopathfinder.naming_conventions.bmon_naming import BMonFilename

logging.basicConfig(level=logging.INFO)


class TestBMonFilename(unittest.TestCase):

    def setUp(self):
        self.var_name = "BMON_DM_ENSEMBLE"
        self.sres = "500m"
        self.timestamp_1 = datetime(2016, 1, 1, 12)
        self.timestamp_2 = datetime(2017, 1, 1, 12)
        self.version = "v1"

        fields_1 = {'var_name': self.var_name, 'sres': self.sres, 'timestamp': self.timestamp_1,
                    'version': self.version}

        self.bmon_fn_1 = BMonFilename(fields_1, convert=True)

        fields_2 = {'var_name': self.var_name, 'sres': self.sres, 'timestamp': self.timestamp_2,
                    'version': self.version}

        self.bmon_fn_2 = BMonFilename(fields_2, convert=True)

        fn = "BMON_DM_ENSEMBLE_500m_20160101120000_v1.nc"
        self.bmon_fn_3 = BMonFilename.from_filename(fn, convert=True)

        self.bmon_fn_4 = BMonFilename.from_filename(fn, convert=False)

    def test_build_bmon_filename(self):
        """
        Test building SGRT file name.

        """
        fn = ("BMON_DM_ENSEMBLE_500m_20160101120000_v1.nc")

        self.assertEqual(str(self.bmon_fn_1), fn)

    def test_get_n_set_date(self):
        """
        Test set and get start and end date.

        """

        self.assertEqual(self.bmon_fn_2['timestamp'], self.timestamp_2)

        new_timestamp = datetime(2009, 1, 1, 12, 23, 33)
        self.bmon_fn_2['timestamp'] = new_timestamp

        self.assertEqual(self.bmon_fn_2['timestamp'], new_timestamp)

    def test_create_bmon_filename(self):
        """
        Tests the creation of a SmartFilename from a given string filename.

        """

        # testing for decoded types
        self.assertEqual(self.bmon_fn_3['var_name'], self.var_name)
        self.assertEqual(self.bmon_fn_3['sres'], '500m')
        self.assertEqual(self.bmon_fn_3['timestamp'], self.timestamp_1)
        self.assertEqual(self.bmon_fn_3['version'], 'v1')
        self.assertEqual(self.bmon_fn_3.ext, '.nc')

        # testing for string types
        self.assertEqual(self.bmon_fn_4['var_name'], self.var_name)
        self.assertEqual(self.bmon_fn_4['sres'], '500m')
        self.assertEqual(self.bmon_fn_4['timestamp'], '20160101120000')
        self.assertEqual(self.bmon_fn_4['version'], 'v1')
        self.assertEqual(self.bmon_fn_4.ext, '.nc')


if __name__ == "__main__":
    unittest.main()




# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import copy
from collections import OrderedDict


class SmartFilenamePart:
    """ Represents a part of filename. """

    def __init__(self, arg, start=0, length=None, delimiter="_", pad="-", decoder=None, encoder=None, compact=False):
        """
        Constructor of SmartFilenamePart class.

        Parameters
        ----------
        arg: object
            Input argument, which can be a string or the decoded part of the filename.
        start: int, optional
            Start index of filename part (default is 0).
        length: int, optional
            Length of filename part.
        delimiter : str, optional
            Delimiter (default: '_')
        pad : str, optional
            Padding symbol (default: '-').
        decoder: function, optional
            Decodes a certain value (str -> object).
        encoder: function, optional
            Encodes a certain value (object -> str).
        """

        self.arg = arg
        self.start = start
        self.delimiter = delimiter
        self.pad = pad
        self.compact = compact
        self.decoder = (lambda x: x) if decoder is None else decoder
        self.encoder = (lambda x: x) if encoder is None else encoder
        length = 0 if compact else length
        self.length = length if length is not None and length != 0 else len(self.encoded)

        # check validity
        if not self.has_valid_len():
            err_msg = "Length does not comply with definition: {:} > {:}".format(len(self), self.length)
            raise ValueError(err_msg)

    def has_valid_len(self):
        """
        Checks if a SmartFilenamePart instance has a valid len.
        It is valid if the specified length is equal to the length of the SmartFilenamePart,
        or if len == 0, i.e accepting any length.

        Returns
        -------
        bool
            True if SmartFilenamePart instance is valid, else False.
        """

        # 0 for accepting any length
        if self.length == 0 or self.compact:
            check = True
        else:
            check = self.length == len(self)

        return check

    @property
    def encoded(self):
        """
        Converts filename part to an encoded (string) representation.

        Returns
        -------
        str
            Encoded (string) representation of a filename part.
        """
        return self.encoder(self.arg)

    @property
    def decoded(self):
        """
        Converts filename part to a decoded (object) representation.

        Returns
        -------
        object
            Decoded (object) representation of a filename part.
        """

        enc_wo_pad = self.encoded.strip(self.pad)
        if enc_wo_pad != '':
            return self.decoder(enc_wo_pad)
        else:
            return None

    def __str__(self):
        """
        Returns the string representation of the field.

        Returns
        -------
        str
            String representation of the class.
        """
        if self.compact and self.arg == "":
            return ''
        return self.encoded.ljust(self.length, self.pad)

    def __repr__(self):
        """
        Returns the class representation, which is the field + delimiter.

        Returns
        -------
        str
            String representation of the class.
        """
        return str(self) + self.delimiter

    def __len__(self):
        """
        Returns length of the filename part.

        Returns
        -------
        int
            Length of the filename part.
        """

        return len(str(self))

    def __add__(self, other):
        """
        Defines summation rule for two SmartFilenamePart/str instances.

        Parameters
        ----------
        other: SmartFilenamePart, str
            Second summand.

        Returns
        -------
        str
            Concatenated strings separated by a delimiter.
        """

        return repr(self) + str(other)


class SmartFilename(object):

    """
    SmartFilename class handles file names with pre-defined field names
    and field length.
    """

    def __init__(self, fields, fields_def, ext=None, pad='-', delimiter='_', convert=False, compact=False):
        """
        Define name of fields, length, pad and delimiter symbol.

        Parameters
        ----------
        fields : dict
            Name of fields (keys) and (values).
        fields_def : OrderedDict
            Name of fields (keys) in right order and length (values). It must contain:
                - "len": int
                    Length of filename part (must be given).
                    "0" to allow any length.
                - "start": int, optional
                    Start index of filename part (default is 0).
                - "delim": str, optional
                    Delimiter between this and the following filename part (default is the one from the parent class).
                - "pad": str,
                    Padding for filename part (default is the one from the parent class).
        ext : str, optional
            File name extension (default: None).
        pad : str, optional
            Padding symbol (default: '-').
        delimiter : str, optional
            Delimiter (default: '_')
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).
        compact: bool, optional
            If true, empty fields are replaced by a single pad character instead of the whole length.
        """
        self.ext = ext
        self.delimiter = delimiter
        self.pad = pad
        self.convert = convert
        self.compact = compact
        self._fn_map = self.__build_map(fields, fields_def)
        self.obj = self.__init_filename_obj()

    @classmethod
    def from_filename(cls, filename_str, fields_def, pad="-", delimiter="_", convert=False, compact=False):
        """
        Converts a filename given as a string into a SmartFilename class object.

        Parameters
        ----------
        filename_str : str
            Filename without any paths (e.g., "M20170725_test.tif").
        fields_def : OrderedDict
            Name of fields (keys) in right order and length (values). It must contain:
                - "len": int
                    Length (positive number) of filename part (must be given).
                    0 to allow any length.
                - "start": int, optional
                    Start index of filename part (default is 0).
                - "delim": str, optional
                    Delimiter between this and the following filename part (default is the one from the parent class).
                - "pad": str,
                    Padding for filename part (default is the one from the parent class).
        pad : str, optional
            Padding symbol (default: '-').
        delimiter : str, optional
            Delimiter (default: '_')
        convert: bool, optional
            If true, decoding is applied to parts of the filename, where such an operation is available (default is False).

        Returns
        -------
        SmartFilename
            Class representing a filename.
        """

        # get extensions from filename
        ext = os.path.splitext(filename_str)[1]

        fields = dict()
        pos = 0
        for name, value in fields_def.items():
            start = value.get('start')
            length = value.get('len')
            delim = value.get('delim')

            if length is not None and length < 0:
                err_msg = "The length of the attribute '{}' must be a positive number".format(name)
                raise ValueError(err_msg)

            # parse part of filename via start and end position
            if start is not None and length is not None:
                fields[name] = filename_str[start:(start + length)]
            elif length is not None:
                start = pos
                if compact:
                    length = 0
                    if delim is None and delimiter is None:
                        raise Exception('The compact filename design requires a delimiter for each field!')

                if length == 0:  # handle variable length
                    end = filename_str.find(delimiter, start) if delimiter in filename_str[start:] \
                        else filename_str.find('.', start)
                    length = end - start
                # if length is not 0, i.e. an entry is available, then add the field. Otherwise skip it.
                if length > 0:
                    fields[name] = filename_str[start:(start + length)]
            else:
                length = 0

            pos += length

            if delim is not None:
                pos += len(delim)
            else:
                pos += len(delimiter)

        if cls.__name__ == "SmartFilename":
            return cls(fields, fields_def, ext=ext, convert=convert, pad=pad, delimiter=delimiter)
        else:
            return cls(fields, ext=ext, convert=convert)

    def __init_filename_obj(self):
        """
        Initialises the class 'FilenameObj' to set all filename attributes as class variables.
        This enables an easier access to filename properties.

        Returns
        -------
        FilenameObj

        """

        class FilenameObj(object):
            def __init__(self):
                pass

        filename_obj = FilenameObj()
        for name, fn_part in self._fn_map.items():
            if self.convert:
                setattr(filename_obj, name, fn_part.decoded)
            else:
                setattr(filename_obj, name, fn_part.encoded)

        return filename_obj

    def __build_map(self, fields, fields_def):
        """
        Creates a dictionary/map between filename part names and SmartFilenamePart instances.

        Paramaters
        ----------
        fields : dict
            Name of fields (keys) and (values).
        fields_def : OrderedDict
            Name of fields (keys) in right order and length (values). It must contain:
                - "len": int
                    Length (positive number) of filename part (must be given).
                    0 to allow any length.
                - "start": int, optional
                    Start index of filename part (default is 0).
                - "delim": str, optional
                    Delimiter between this and the following filename part (default is the one from the parent class).
                - "pad": str,
                    Padding for filename part (default is the one from the parent class).

        Returns
        -------
        dict
            Contains SmartFilenamePart instances representing each part of the filename.
        """

        # check fields consistency
        for key in fields.keys():
            if key not in fields_def.keys():
                raise KeyError("Field name undefined: {:}".format(key))

        fn_map = OrderedDict()
        last_key_name = list(fields_def.keys())[-1]
        for name, keys in fields_def.items():
            if name not in fields:
                elem = ""
            else:
                elem = fields[name]

            fn_part_kwargs = dict()
            if 'delim' not in keys:
                fn_part_kwargs['delimiter'] = self.delimiter
            else:
                fn_part_kwargs['delimiter'] = keys['delim']

            if 'len' in keys:
                # check delimiter in case of zero length
                if keys['len'] == 0 and fn_part_kwargs['delimiter'] == '':
                    err_msg = 'A variable field length (length = 0) requires a delimiter!'
                    raise ValueError(err_msg)

                if keys['len'] < 0:
                    err_msg = "The length of the attribute '{}' must be a positive number".format(name)
                    raise ValueError(err_msg)

                fn_part_kwargs['length'] = keys['len']

            if 'decoder' in keys:
                fn_part_kwargs['decoder'] = keys['decoder']
            if 'encoder' in keys:
                fn_part_kwargs['encoder'] = keys['encoder']
            fn_part_kwargs['compact'] = self.compact

            if name == last_key_name:  # set empty delimiter for last field
                fn_part_kwargs['delimiter'] = ''

            # reset delimiter of last element to be empty
            smart_fn_part = SmartFilenamePart(elem, **fn_part_kwargs)
            fn_map[name] = smart_fn_part

        return fn_map

    def _build_fn(self):
        """
        Build file name based on fields, padding and length.

        Returns
        -------
        filename : str
            Filled file name.

        """

        fn_parts = list(self._fn_map.values())

        filename = ''.join([repr(fn_part) for fn_part in fn_parts])

        # handle case if last field is not defined (skip the last delimeter then)
        if filename.endswith(self.delimiter):
            filename = filename[:-1]

        if self.ext is not None:
            filename += self.ext

        return filename

    def _get_field(self, name):
        """
        Returns the value of the field with a given key.

        Parameters
        ----------
        name : str
            Name of the field.

        Returns
        -------
        str, object
            Part of the filename associated with given key. Depending on the chosen flag 'convert', it is either a str
            (convert=False) or an object.
        """

        # check and reset the attribute of the object variable
        field_from_obj = self._fn_map[name].encoder(getattr(self.obj, name))
        if field_from_obj and (field_from_obj != self._fn_map[name].encoded):
            fn_part = copy.deepcopy(self._fn_map[name])
            fn_part.arg = field_from_obj
            if fn_part.has_valid_len():
                self._fn_map[name] = fn_part

        if self.convert:
            return self._fn_map[name].decoded
        else:
            return self._fn_map[name].encoded.strip(self._fn_map[name].pad)

    def __getitem__(self, name):
        """
        Returns the value of the field with a given key.

        Parameters
        ----------
        name : str
            Name of the field.

        Returns
        -------
        str, object
            Part of the filename associated with given key. Depending on the chosen flag 'convert', it is either a str
            (convert=False) or an object. If the key can't be found in the fields definition, the method tries to return
            a property of an inherited class.
        """

        if name in self._fn_map:
            return self._get_field(name)
        elif hasattr(self, name):
            return getattr(self, name)
        else:
            raise KeyError('"{}" is neither a class variable nor a file attribute.'.format(name))

    def __setitem__(self, name, value):
        """
        Sets the value of a filename field corresponding to the given key.

        Parameters
        ----------
        name : str
            Name of the field.
        value: object
            Value of the field.

        """

        if name in self._fn_map:
            fn_part = copy.deepcopy(self._fn_map[name])
            fn_part.arg = value
            if self.compact:
                fn_part.length = 0
            if not fn_part.has_valid_len():
                err_msg = "Length does not comply with definition: {:} > {:}".format(len(fn_part.encoded),
                                                                                     fn_part.length)
                raise ValueError(err_msg)
            else:
                self._fn_map[name] = fn_part
                value = fn_part.encoded.replace(self.pad, '')
                if self.convert:
                    setattr(self.obj, name, fn_part.decoded)
                else:
                    setattr(self.obj, name, value)
        else:
            raise KeyError("Field name undefined: {:}".format(name))

    def __repr__(self):
        """
        Returns the string representation of the class.

        Returns
        -------
        str
            String representation of the class.
        """

        return self._build_fn()
